from importlib import reload
from unittest import mock

import pytest
from pkgcore.ebuild import eapi
from pkgcore.ebuild.eapi import EAPI, eapi6, get_eapi


def test_get_eapi():
    # unknown EAPI
    unknown_eapi = get_eapi("unknown")
    assert unknown_eapi in EAPI.unknown_eapis.values()
    # check that unknown EAPI is now registered as an unknown
    assert unknown_eapi == get_eapi("unknown")

    # known EAPI
    eapi = get_eapi("6")
    assert eapi6 == eapi


class TestEAPI:

    def test_register(self):
        # re-register known EAPI
        with pytest.raises(ValueError):
            EAPI.register(magic="0")

        with mock.patch('pkgcore.ebuild.eapi.bash_version') as bash_version, \
                mock.patch.dict(eapi.EAPI.known_eapis):
            # inadequate bash version
            bash_version.return_value = '3.1'
            with pytest.raises(SystemExit) as excinfo:
                new_eapi = EAPI.register(magic='new', optionals={'bash_compat': '3.2'})
            assert "EAPI 'new' requires >=bash-3.2, system version: 3.1" == excinfo.value.args[0]

            # adequate system bash versions
            bash_version.return_value = '3.2'
            test_eapi = EAPI.register(magic='test', optionals={'bash_compat': '3.2'})
            assert test_eapi._magic == 'test'
            bash_version.return_value = '4.2'
            test_eapi = EAPI.register(magic='test1', optionals={'bash_compat': '4.1'})
            assert test_eapi._magic == 'test1'

    def test_is_supported(self, caplog):
        assert eapi6.is_supported

        with mock.patch.dict(eapi.EAPI.known_eapis):
            # partially supported EAPI is flagged as such
            test_eapi = EAPI.register("test", optionals={'is_supported': False})
            assert test_eapi.is_supported
            assert caplog.text.endswith("EAPI 'test' isn't fully supported\n")

            # unsupported/unknown EAPI is flagged as such
            unknown_eapi = get_eapi("blah")
            assert not unknown_eapi.is_supported

    def test_inherits(self):
        for eapi_str, eapi_obj in EAPI.known_eapis.items():
            objs = (get_eapi(str(x)) for x in range(int(eapi_str), -1, -1))
            assert list(map(str, eapi_obj.inherits)) == list(map(str, objs))

    def test_ebd_env(self):
        for eapi_str, eapi_obj in EAPI.known_eapis.items():
            assert eapi_obj.ebd_env['EAPI'] == eapi_str



__all__ = (
    "null_output", "formatter_output", "file_handle_output",
    "phase_observer", "repo_observer", "decorate_build_method",
)

import threading

from snakeoil import klass
from snakeoil.currying import pre_curry


def _convert(msg, args=(), kwds={}):
    # Note for interpolation, ValueError can be thrown by '%2(s'
    # TypeError by "%i" % "2", and KeyError via what you would expect.
    if args:
        if kwds:
            raise TypeError(
                "both position and optional args cannot be "
                "supplied: given msg(%r), args(%r), kwds(%r)"
                % (msg, args, kwds))
        try:
            return msg % args
        except (ValueError, TypeError) as e:
            raise TypeError(
                f"observer interpolation error: {e}, msg={msg!r}, args={args!r}")
    elif kwds:
        try:
            return msg % kwds
        except (KeyError, TypeError, ValueError) as e:
            raise TypeError(
                f"observer interpolation error: {e}, msg={msg!r}, kwds={kwds!r}")
    return msg


class null_output:

    def warn(self, msg, *args, **kwds):
        pass

    def error(self, msg, *args, **kwds):
        pass

    def info(self, msg, *args, **kwds):
        pass

    def debug(self, msg, *args, **kwds):
        pass

    def write(self, msg, *args, **kwds):
        pass

    def flush(self):
        pass


class formatter_output(null_output):

    def __init__(self, out):
        self._out = out
        self.verbosity = getattr(out, 'verbosity', 0)

    def debug(self, msg, *args, **kwds):
        self._out.write(_convert("debug: " + msg, args, kwds))

    def error(self, msg, *args, **kwds):
        prefixes = kwds.pop(
            'prefixes', (self._out.fg('red'), self._out.bold, ' * ', self._out.reset))
        self._out.write(_convert(msg, args, kwds), prefixes=prefixes)

    def info(self, msg, *args, **kwds):
        prefixes = kwds.pop(
            'prefixes', (self._out.fg('green'), self._out.bold, ' * ', self._out.reset))
        self._out.write(_convert(msg, args, kwds), prefixes=prefixes)

    def warn(self, msg, *args, **kwds):
        prefixes = kwds.pop(
            'prefixes', (self._out.fg('yellow'), self._out.bold, ' * ', self._out.reset))
        self._out.write(_convert(msg, args, kwds), prefixes=prefixes)

    def write(self, msg, *args, autoline=False, **kwds):
        self._out.write(_convert(msg, args, kwds), autoline=autoline)

    def flush(self):
        self._out.flush()


class file_handle_output(formatter_output):

    def debug(self, msg, *args, **kwds):
        self._out.write(f"debug: {_convert(msg, args, kwds)}\n")

    def error(self, msg, *args, **kwds):
        self._out.write(f"error: {_convert(msg, args, kwds)}\n")

    def info(self, msg, *args, **kwds):
        self._out.write(f"info: {_convert(msg, args, kwds)}\n")

    def warn(self, msg, *args, **kwds):
        self._out.write(f"warning: {_convert(msg, args, kwds)}\n")

    def write(self, msg, *args, **kwds):
        self._out.write(_convert(msg, args, kwds))


class phase_observer:

    def __init__(self, output, debug=False):
        self._output = output
        self.verbosity = getattr(output, 'verbosity', 0)
        self._debug = debug

    def phase_start(self, phase):
        if self._debug:
            self._output.write(f"starting {phase}\n")

    def debug(self, msg, *args, **kwds):
        if self._debug:
            self._output.debug(msg, *args, **kwds)

    info = klass.alias_attr("_output.info")
    warn = klass.alias_attr("_output.warn")
    error = klass.alias_attr("_output.error")
    write = klass.alias_attr("_output.write")
    flush = klass.alias_attr("_output.flush")

    def phase_end(self, phase, status):
        if self._debug:
            self._output.write(f"finished {phase}: {status}\n")


class repo_observer(phase_observer):

    def trigger_start(self, hook, trigger):
        if self._debug:
            self._output.write(f"hook {hook}: trigger: starting {trigger!r}\n", hook)

    def trigger_end(self, hook, trigger):
        if self._debug:
            self._output.write(f"hook {hook}: trigger: finished {trigger!r}\n", hook)

    def installing_fs_obj(self, obj):
        self._output.write(f">>> {obj}\n")

    def removing_fs_obj(self, obj):
        self._output.write(f"<<< {obj}\n")


def _reflection_func(attr, self, *args, **kwds):
    return self._invoke(attr, *args, **kwds)


def _mk_observer_proxy(target):
    class foo(target):
        for x in set(dir(target)).difference(dir(object)):
            locals()[x] = pre_curry(_reflection_func, x)
    return foo


class threadsafe_repo_observer(_mk_observer_proxy(repo_observer)):

    def __init__(self, observer):
        self._observer = observer
        self._lock = threading.Lock()

    def _invoke(self, attr, *args, **kwds):
        self._lock.acquire()
        try:
            return getattr(self._observer, attr)(*args, **kwds)
        finally:
            self._lock.release()


def wrap_build_method(phase, method, self, *args, **kwds):
    disable_observer = kwds.pop("disable_observer", False)
    if not hasattr(self.observer, 'phase_start') or disable_observer:
        return method(self, *args, **kwds)
    self.observer.phase_start(phase)
    ret = False
    try:
        ret = method(self, *args, **kwds)
    finally:
        self.observer.phase_end(phase, ret)
    return ret


def decorate_build_method(phase):
    def f(func):
        return pre_curry(wrap_build_method, phase, func)
    return f



"""
WARNING: this module is for testing usage only; it disables
the default strictness tests that fsBase.* derivatives have to ease
testing.  Do not use it in non-test code.
"""

from pkgcore.fs import fs
# we use pre_curry to preserve the docs for the wrapped target
from snakeoil.currying import pre_curry

# we're anal about del'ing here to prevent the vars from lingering around,
# showing up when folks are poking around

key = None
for key in dir(fs):
    val = getattr(fs, key)
    # protection; issubclass pukes if it's not a class.
    # downside, this works on new style only
    if isinstance(val, type) and issubclass(val, fs.fsBase) and \
            val is not fs.fsBase:
        locals()[f"_original_{key}"] = val
        val = pre_curry(val, strict=False)
        val.__doc__ = locals()[f"_original_{key}"].__doc__
    locals()[key] = val
    del val

del key



__all__ = (
    "SyncError", "UriError", "MissingLocalUser", "MissingBinary",
    "Syncer", "ExternalSyncer", "VcsSyncer",
    "GenericSyncer", "DisabledSyncer", "AutodetectSyncer",
)

import os
import pwd
import stat

from snakeoil import process

from .. import os_data, plugin
from ..config.hint import ConfigHint, configurable
from ..exceptions import PkgcoreUserException


class SyncError(PkgcoreUserException):
    """Generic syncing error."""


class UriError(SyncError):

    def __init__(self, uri, msg):
        self.uri = uri
        self.msg = msg
        super().__init__(f"{msg}: {uri!r}")


class PathError(SyncError):

    def __init__(self, path, msg):
        self.path = path.rstrip(os.path.sep)
        self.msg = msg
        super().__init__(f"{msg}: {self.path!r}")


class MissingLocalUser(SyncError):

    def __init__(self, uri, msg):
        self.uri = uri
        self.msg = msg
        super().__init__(f"{msg}: {uri!r}")


class MissingBinary(SyncError):

    def __init__(self, binary, msg):
        self.binary = binary
        self.msg = msg
        super().__init__(f"{msg}: {binary!r}")


class Syncer:

    forcable = False

    supported_uris = ()
    supported_protocols = ()
    supported_exts = ()

    # plugin system uses this.
    disabled = False

    pkgcore_config_type = ConfigHint(
        {'path': 'str', 'uri': 'str', 'opts': 'str', 'usersync': 'bool'},
        typename='syncer')

    def __init__(self, path, uri, default_verbosity=0, usersync=False, opts=''):
        self.verbosity = default_verbosity
        self.usersync = usersync
        self.basedir = path.rstrip(os.path.sep) + os.path.sep
        uri = self.parse_uri(uri)
        self.uid, self.gid, self.uri = self.split_users(uri)
        self.opts = opts.split()

    @staticmethod
    def parse_uri(uri):
        """Return the real URI with any protocol prefix stripped."""
        return uri

    @classmethod
    def is_usable_on_filepath(cls, path):
        return None

    def split_users(self, raw_uri):
        """
        :param raw_uri: string uri to split users from; harring::ferringb:pass
          for example is local user 'harring', remote 'ferringb',
          password 'pass'
        :return: (local user, remote user, remote pass), defaults to the
            current process's uid if no local user specified
        """
        uri = raw_uri.split("::", 1)
        if len(uri) == 1:
            if self.usersync:
                if os.path.exists(self.basedir):
                    stat = os.stat(self.basedir)
                    uid = stat.st_uid
                    gid = stat.st_gid
                else:
                    uid = os_data.portage_uid
                    gid = os_data.portage_gid
            else:
                uid = os_data.uid
                gid = os_data.gid

            return uid, gid, raw_uri
        try:
            if uri[1].startswith("@"):
                uri[1] = uri[1][1:]
            if '/' in uri[0] or ':' in uri[0]:
                proto = uri[0].split("/", 1)
                proto[1] = proto[1].lstrip("/")
                uri[0] = proto[1]
                uri[1] = f"{proto[0]}//{uri[1]}"

            return pwd.getpwnam(uri[0]).pw_uid, os_data.gid, uri[1]
        except KeyError as e:
            raise MissingLocalUser(raw_uri, str(e))

    def sync(self, verbosity=None, force=False):
        if self.disabled:
            return False
        kwds = {}
        if self.forcable and force:
            kwds["force"] = True
        if verbosity is None:
            verbosity = self.verbosity
        # output_fd is harded coded as stdout atm.
        return self._sync(verbosity, 1, **kwds)

    def _sync(self, verbosity, output_fd, **kwds):
        raise NotImplementedError(self, "_sync")

    def __str__(self):
        return f"{self.__class__} syncer: {self.basedir}, {self.uri}"

    @classmethod
    def supports_uri(cls, uri):
        for prefix, level in cls.supported_uris:
            if uri.startswith(prefix):
                return level
        if uri.startswith(cls.supported_protocols) and uri.endswith(cls.supported_exts):
            return 1
        return 0


class ExternalSyncer(Syncer):

    """Base class for syncers that spawn a binary to do the the actual work."""

    binary = None

    # external env settings passed through to syncing commands
    env_whitelist = (
        'SSH_AUTH_SOCK',
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.env = {v: os.environ[v] for v in self.env_whitelist if v in os.environ}

        if not hasattr(self, 'binary_path'):
            self.binary_path = self.require_binary(self.binary)

    @staticmethod
    def require_binary(bin_name, fatal=True):
        try:
            return process.find_binary(bin_name)
        except process.CommandNotFound as e:
            if fatal:
                raise MissingBinary(bin_name, str(e))
            return None

    @classmethod
    def _plugin_disabled_check(cls):
        disabled = getattr(cls, '_disabled', None)
        if disabled is None:
            path = getattr(cls, 'binary_path', None)
            if path is None:
                if cls.binary is None:
                    disabled = cls._disabled = True
                else:
                    disabled = cls._disabled = (
                        cls.require_binary(cls.binary, fatal=False) is None)
            else:
                disabled = cls._disabled = os.path.exists(path)
        return disabled

    def _spawn(self, command, pipes, **kwargs):
        return process.spawn.spawn(
            command, fd_pipes=pipes, uid=self.uid, gid=self.gid, env=self.env, **kwargs)

    @staticmethod
    def _rewrite_uri_from_stat(path, uri):
        chunks = uri.split("//", 1)
        if len(chunks) == 1:
            return uri
        try:
            return "%s//%s::%s" % (
                chunks[0], pwd.getpwuid(os.stat(path).st_uid)[0], chunks[1])
        except KeyError:
            # invalid uid, reuse the uri
            return uri


class VcsSyncer(ExternalSyncer):

    def _sync(self, verbosity, output_fd):
        try:
            st = os.stat(self.basedir)
        except FileNotFoundError:
            command = self._initial_pull() + self.opts
            chdir = None
        except EnvironmentError as e:
            raise PathError(self.basedir, e.strerror) from e
        else:
            if not stat.S_ISDIR(st.st_mode):
                raise PathError(self.basedir, "isn't a directory")
            command = self._update_existing() + self.opts
            chdir = self.basedir

        # we assume syncers support -v and -q for verbose and quiet output
        if verbosity < 0:
            command.append('-q')
        elif verbosity > 0:
            command.append('-' + 'v' * verbosity)

        ret = self._spawn(command, pipes={1: output_fd, 2: output_fd, 0: 0},
                          cwd=chdir)
        return ret == 0

    def _initial_pull(self):
        raise NotImplementedError(self, "_initial_pull")

    def _update_existing(self):
        raise NotImplementedError(self, "_update_existing")


@configurable(
    {'basedir': 'str', 'uri': 'str', 'usersync': 'bool', 'opts': 'str'},
    typename='syncer')
def GenericSyncer(basedir, uri, **kwargs):
    """Syncer using the plugin system to find a syncer based on uri."""
    plugins = list(
        (plug.supports_uri(uri), plug)
        for plug in plugin.get_plugins('syncer'))
    plugins.sort(key=lambda x: x[0])
    if not plugins or plugins[-1][0] <= 0:
        raise UriError(uri, "no known syncer support")
    # XXX this is random if there is a tie. Should we raise an exception?
    return plugins[-1][1](basedir, uri, **kwargs)


class DisabledSyncer(Syncer):

    disabled = True

    def __init__(self, path, *args, **kwargs):
        super().__init__(path, uri='')


@configurable({'basedir': 'str', 'usersync': 'bool'}, typename='syncer')
def DisabledSync(basedir, *args, **kwargs):
    return DisabledSyncer(basedir)


@configurable({'basedir': 'str', 'usersync': 'bool'}, typename='syncer')
def AutodetectSyncer(basedir, **kwargs):
    for syncer_cls in plugin.get_plugins('syncer'):
        if args := syncer_cls.is_usable_on_filepath(basedir):
            return syncer_cls(basedir, *args, **kwargs)
    return DisabledSyncer(basedir, **kwargs)



"""pmerge formatting module"""

__all__ = (
    "Formatter", "use_expand_filter",
    "BasicFormatter", "PkgcoreFormatter", "CountingFormatter",
    "PortageFormatter", "PortageVerboseFormatter",
)

import operator
import os

from snakeoil.cli.input import userquery
from snakeoil.mappings import defaultdictkey
from snakeoil.osutils import pjoin, sizeof_fmt
from snakeoil.sequences import iflatten_instance
from snakeoil.strings import pluralism

from ..config.hint import ConfigHint
from ..log import logger


class use_expand_filter:

    def __init__(self, use_expand, use_expand_hidden):
        """
        :type use_expand: iterable of strings
        :param use_expand: names of use-expanded variables.
        :type use_expand_hidden: set of strings
        :param use_expand_hidden: names of use-expanded vars that should not
            be added to the dict.
        """
        self.expand_filters = {x.lower(): (x not in use_expand_hidden, x)
                               for x in use_expand}
        self.use_expand = use_expand
        self.use_expand_hidden = use_expand_hidden
        self.known_flags = {}

    def __call__(self, use):
        """Split USE flags up into "normal" flags and use-expanded ones.
        :type use: iterable of strings
        :param use: flags that are set.
        :rtype: sequence of strings, dict mapping a string to a list of strings
        :return: set of normal flags and a mapping from use_expand name to
            value (with the use-expanded bit stripped off, so
            C{"video_cards_alsa"} becomes C{"{'video_cards': ['alsa']}"}).
        """
        # XXX: note this is fairly slow- actually takes up more time then
        # chunks of the resolver
        ue_dict = {}
        usel = []
        ef = self.expand_filters
        kf = self.known_flags

        for flag in use:
            data = kf.get(flag)
            if data is None:
                split_flag = flag.rsplit("_", 1)
                while len(split_flag) == 2:
                    if split_flag[0] not in ef:
                        split_flag = split_flag[0].rsplit("_", 1)
                        continue
                    expand_state = ef[split_flag[0]]
                    if expand_state[0]:
                        # not hidden
                        kf[flag] = data = (expand_state[1], flag[len(split_flag[0]) + 1:])
                    else:
                        kf[flag] = data = False
                    break
                else:
                    kf[flag] = data = True
            if data is True:
                # straight use flag.
                usel.append(flag)
            elif data:
                # non hidden flag.
                if not data[0] in ue_dict:
                    ue_dict[data[0]] = set([data[1]])
                else:
                    ue_dict[data[0]].add(data[1])

        return frozenset(usel), ue_dict


class Formatter:
    """Base Formatter class: All formatters should be subclasses of this."""

    pkgcore_config_type = ConfigHint(typename='pmerge_formatter', raw_class=True)

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def format(self, op):
        """Formats an op. Subclasses must define this method"""
        raise NotImplementedError(self.format)

    def ask(self, question, responses=None, default_answer=None, limit=3):
        return userquery(
            question, self.out, self.err, responses, default_answer, limit)

    def end(self):
        """Called at the end, normally for summary information"""


class BasicFormatter(Formatter):
    """A basic formatter, intended for scripts"""

    def format(self, op):
        self.out.write(op.pkg.key)


class VerboseFormatter(Formatter):
    """Formatter with output forced into verbose mode."""

    def __init__(self, **kwargs):
        kwargs['verbosity'] = 1
        super().__init__(**kwargs)


class PkgcoreFormatter(Formatter):
    """The original pkgcore output"""

    def format(self, op):
        repo = getattr(op.pkg.repo, 'repo_id', None)
        if not repo:
            p = str(op.pkg.cpvstr)
        else:
            p = f"{op.pkg.cpvstr}::{repo}"
        if op.desc == "replace":
            self.out.write(f"replace {op.old_pkg.cpvstr}, {p}")
        else:
            self.out.write(f"{op.desc.ljust(7)} {p}")


class CountingFormatter(Formatter):
    """Subclass for formatters that count packages"""

    def __init__(self, **kwargs):
        kwargs.setdefault("verbosity", 0)
        super().__init__(**kwargs)
        self.package_data = defaultdictkey(lambda x: 0)

        # total download size for all pkgs to be merged
        self.download_size = 0

    def visit_op(self, op_type):
        """Track the number of package operations of each type."""
        self.package_data[op_type] += 1

    def end(self):
        """Output total package, operation, and download size counts."""
        self.out.write()
        if self.verbosity > 0:
            total = sum(self.package_data.values())
            self.out.write(
                f"Total: {total} package{pluralism(total)}", autoline=False)

            d = dict(self.package_data.items())
            op_types = (
                ('add', 'new'),
                ('upgrade', 'upgrade'),
                ('downgrade', 'downgrade'),
                ('slotted_add', 'in new slot'),
                ('replace', 'reinstall'),
            )
            op_list = []
            for op_type, op_str in op_types:
                num_ops = d.pop(op_type, 0)
                if num_ops:
                    if op_str == 'new':
                        op_list.append(f"{num_ops} {op_str}")
                    else:
                        op_list.append(f"{num_ops} {op_str}{pluralism(num_ops)}")
            if d:
                op_list.append(f"{len(d)} other op{pluralism(d)}")
            if op_list:
                self.out.write(f" ({', '.join(op_list)})", autoline=False)
            if self.download_size:
                self.out.write(
                    ', Size of downloads: ', sizeof_fmt(self.download_size), autoline=False)
            self.out.write()


class PortageFormatter(CountingFormatter):
    """Formatter designed to resemble portage output."""

    def __init__(self, **kwargs):
        kwargs.setdefault("use_expand", set())
        kwargs.setdefault("use_expand_hidden", set())
        super().__init__(**kwargs)
        self.use_splitter = use_expand_filter(
            self.use_expand, self.use_expand_hidden)
        # Map repo location to an index.
        self.repos = {}
        # set of files to be downloaded
        self.downloads = set()

    def format(self, op):
        # <type>       - ebuild, block or nomerge (for --tree)
        #       N      - new package
        #        R     - rebuild package
        #         F    - fetch restricted
        #         f    - fetch restricted already downloaded
        #          D   - downgrade
        #           U  - updating to another version
        #            # - masked
        #            * - missing keyword
        #            ~ - unstable keyword
        # Caveats:
        # - U and D are both displayed to show a downgrade - this is kept
        # in order to be consistent with existing portage behaviour

        out = self.out
        origautoline = out.autoline
        out.autoline = False

        self.pkg_disabled_use = self.pkg_forced_use = set()
        if hasattr(self, 'pkg_get_use'):
            self.pkg_forced_use, _, self.pkg_disabled_use = self.pkg_get_use(op.pkg)

        # This is for the summary at the end
        if self.quiet_repo_display:
            self.repos.setdefault(op.pkg.repo, len(self.repos)+1)

        pkg_is_bold = any(x.match(op.pkg) for x in getattr(self, 'world_list', ()))

        # We don't do blockers or --tree stuff yet
        data = ['[']
        pkg_coloring = []
        if pkg_is_bold:
            pkg_coloring.append(out.bold)
        if op.desc == 'remove':
            pkg_coloring.insert(0, out.fg('red'))
            data += pkg_coloring + ['uninstall']
        elif getattr(op.pkg, 'built', False):
            pkg_coloring.insert(0, out.fg('magenta'))
            data += pkg_coloring + ['binary']
        else:
            pkg_coloring.insert(0, out.fg('green'))
            data += pkg_coloring + ['ebuild']

        data += [out.reset, ' ']
        out.write(*data)

        # Order is important here - look at the above diagram
        op_type = op.desc
        op_chars = [[' '] for x in range(7)]
        if 'fetch' in op.pkg.restrict:
            if all(os.path.isfile(pjoin(self.distdir, f))
                   for f in op.pkg.distfiles):
                fetched = [out.fg('green'), out.bold, 'f', out.reset]
            else:
                fetched = [out.fg('red'), out.bold, 'F', out.reset]
            op_chars[3] = fetched

        if op.desc == "add":
            op_chars[1] = [out.fg('green'), out.bold, 'N', out.reset]
            if op.pkg.slot != '0' and self.installed_repos.match(op.pkg.unversioned_atom):
                op_chars[2] = [out.fg('green'), out.bold, 'S', out.reset]
                op_type = 'slotted_add'
        elif op.desc == "replace":
            if op.pkg == op.old_pkg:
                op_chars[2] = [out.fg('yellow'), out.bold, 'R', out.reset]
            else:
                op_chars[4] = [out.fg('cyan'), out.bold, 'U', out.reset]
                if op.pkg > op.old_pkg:
                    op_type = 'upgrade'
                else:
                    op_chars[5] = [out.fg('blue'), out.bold, 'D', out.reset]
                    op_type = 'downgrade'
        elif op.desc == 'remove':
            pass
        else:
            logger.warning("unformattable op type: desc(%r), %r", op.desc, op)

        if self.verbosity > 0:
            if (self.unstable_arch in op.pkg.keywords and
                    self.unstable_arch not in op.pkg.repo.domain_settings['ACCEPT_KEYWORDS']):
                op_chars[6] = [out.fg('yellow'), out.bold, '~', out.reset]
            elif not op.pkg.keywords:
                op_chars[6] = [out.fg('red'), out.bold, '*', out.reset]
            else:
                if op.pkg.repo.masked.match(op.pkg.versioned_atom):
                    op_chars[6] = [out.fg('red'), out.bold, '#', out.reset]

        out.write(*(iflatten_instance(op_chars)))
        out.write('] ')

        self.visit_op(op_type)

        pkg = [op.pkg.cpvstr]
        if self.verbosity > 0:
            if op.pkg.subslot != op.pkg.slot:
                pkg.append(f":{op.pkg.slot}/{op.pkg.subslot}")
            elif op.pkg.slot != '0':
                pkg.append(f":{op.pkg.slot}")
            if not self.quiet_repo_display and op.pkg.source_repository:
                pkg.append(f"::{op.pkg.source_repository}")
        out.write(*(pkg_coloring + pkg + [out.reset]))

        installed = []
        if op.desc == 'replace':
            old_pkg = [op.old_pkg.fullver]
            if self.verbosity > 0:
                if op.old_pkg.subslot != op.old_pkg.slot:
                    old_pkg.append(f":{op.old_pkg.slot}/{op.old_pkg.subslot}")
                elif op.old_pkg.slot != '0':
                    old_pkg.append(f":{op.old_pkg.slot}")
                if not self.quiet_repo_display and op.old_pkg.source_repository:
                    old_pkg.append(f"::{op.old_pkg.source_repository}")
            if op_type != 'replace' or op.pkg.source_repository != op.old_pkg.source_repository:
                installed = ''.join(old_pkg)
        elif op_type == 'slotted_add':
            if self.verbosity > 0:
                pkgs = sorted(
                    f"{x.fullver}:{x.slot}" for x in
                    self.installed_repos.match(op.pkg.unversioned_atom))
            else:
                pkgs = sorted(
                    x.fullver for x in
                    self.installed_repos.match(op.pkg.unversioned_atom))
            installed = ', '.join(pkgs)

        # output currently installed versions
        if installed:
            out.write(' ', out.fg('blue'), out.bold, f'[{installed}]', out.reset)

        # Build a list of (useflags, use_expand_dicts) tuples.
        # HACK: if we are in "replace" mode we build a list of length
        # 4, else this is a list of length 2. We then pass this to
        # format_use which can take either 2 or 4 arguments.
        uses = ((), ())
        if op.desc == 'replace':
            uses = (
                op.pkg.iuse_stripped, op.pkg.use,
                op.old_pkg.iuse_stripped, op.old_pkg.use)
        elif op.desc == 'add':
            uses = (op.pkg.iuse_stripped, op.pkg.use)
        stuff = list(map(self.use_splitter, uses))

        # Convert the list of tuples to a list of lists and a list of
        # dicts (both length 2 or 4).
        uselists, usedicts = list(zip(*stuff))

        # output USE flags
        self.format_use('use', *uselists)

        # output USE_EXPAND flags
        for expand in sorted(self.use_expand - self.use_expand_hidden):
            try:
                sorter = op.pkg.repo.use_expand_sorter(expand)
            except AttributeError:
                # TODO: hack for unit tests using fake repo objs
                sorter = lambda k: k
            flaglists = [sorted(d.get(expand, ()), key=sorter) for d in usedicts]
            self.format_use(expand, *flaglists, sorter=sorter)

        # output download size
        if self.verbosity > 0:
            if not op.pkg.built:
                downloads = set(
                    f for f in op.pkg.distfiles
                    if not os.path.isfile(pjoin(self.distdir, f)))
                if downloads.difference(self.downloads):
                    self.downloads.update(downloads)
                    size = sum(
                        v['size'] for dist, v in
                        op.pkg.manifest.distfiles.items() if dist in downloads)
                    if size:
                        self.download_size += size
                        out.write(' ', sizeof_fmt(size))

            if self.quiet_repo_display:
                out.write(out.fg('cyan'), f" [{self.repos[op.pkg.repo]}]")

        out.write('\n')
        out.autoline = origautoline

    def format_use(self, attr, pkg_iuse, pkg_use, old_pkg_iuse=None, old_pkg_use=None, sorter=lambda k: k):
        """Write the current selection from a set of flags to a formatter.

        :type attr: string
        :param attr: name of the setting
        :type pkg_iuse: set of strings
        :param pkg_iuse: all available use flags for the package
        :type pkg_use: set of strings
        :param pkg_use: enabled use flags for the package
        :type old_pkg_iuse: set of strings
        :param old_pkg_iuse: all available use flags in the previous version
        :type old_pkg_use: set of strings
        :param old_pkg_use: enabled use flags in the previous version
        """
        out = self.out
        red = out.fg('red')
        green = out.fg('green')
        blue = out.fg('blue')
        yellow = out.fg('yellow')
        bold = out.bold
        reset = out.reset

        flags = []
        enabled = set(pkg_iuse) & set(pkg_use)
        disabled = set(pkg_iuse) - set(pkg_use)

        # updating or rebuilding pkg
        if old_pkg_iuse is not None and old_pkg_use is not None:

            old_enabled = set(old_pkg_iuse) & set(old_pkg_use)
            old_disabled = set(old_pkg_iuse) - set(old_pkg_use)
            removed = set(old_pkg_iuse) - set(pkg_iuse)

            for flag in sorted(enabled, key=sorter):
                expanded_flag = '_'.join((attr.lower(), flag)) if attr != 'use' else flag
                if flag in old_enabled:
                    # unchanged
                    if self.verbosity > 0:
                        if expanded_flag in self.pkg_forced_use:
                            flags.extend(('(', red, bold, flag, reset, ')', ' '))
                        else:
                            flags.extend((red, bold, flag, reset, ' '))
                elif flag in old_disabled:
                    # toggled
                    if expanded_flag in self.pkg_forced_use:
                        flags.extend(('(', green, bold, flag, reset, '*)', ' '))
                    else:
                        flags.extend((green, bold, flag, reset, '*', ' '))
                else:
                    # new
                    if expanded_flag in self.pkg_forced_use:
                        flags.extend(('(', yellow, bold, flag, reset, '%*)', ' '))
                    else:
                        flags.extend((yellow, bold, flag, reset, '%*', ' '))

            for flag in sorted(disabled, key=sorter):
                expanded_flag = '_'.join((attr.lower(), flag)) if attr != 'use' else flag
                if flag in old_disabled:
                    # unchanged
                    if self.verbosity > 0:
                        if expanded_flag in self.pkg_disabled_use:
                            flags.extend(('(', blue, bold, '-', flag, reset, ')', ' '))
                        else:
                            flags.extend((blue, bold, '-', flag, reset, ' '))
                elif flag in old_enabled:
                    # toggled
                    if expanded_flag in self.pkg_disabled_use:
                        flags.extend(('(', green, bold, '-', flag, reset, '*)', ' '))
                    else:
                        flags.extend((green, bold, '-', flag, reset, '*', ' '))
                else:
                    # new
                    if expanded_flag in self.pkg_disabled_use:
                        flags.extend(('(', yellow, bold, '-', flag, reset, '%)', ' '))
                    else:
                        flags.extend((yellow, bold, '-', flag, reset, '%', ' '))

            if self.verbosity > 0:
                for flag in sorted(removed, key=sorter):
                    if flag in old_enabled:
                        flags.extend(('(', yellow, bold, '-', flag, reset, '%*)', ' '))
                    else:
                        flags.extend(('(', yellow, bold, '-', flag, reset, '%)', ' '))

        # new pkg install
        else:
            for flag in sorted(enabled, key=sorter):
                expanded_flag = '_'.join((attr.lower(), flag)) if attr != 'use' else flag
                if expanded_flag in self.pkg_forced_use:
                    flags.extend(('(', red, bold, flag, reset, ')', ' '))
                else:
                    flags.extend((red, bold, flag, reset, ' '))
            for flag in sorted(disabled, key=sorter):
                expanded_flag = '_'.join((attr.lower(), flag)) if attr != 'use' else flag
                if expanded_flag in self.pkg_disabled_use:
                    flags.extend(('(', blue, bold, '-', flag, reset, ')', ' '))
                else:
                    flags.extend((blue, bold, '-', flag, reset, ' '))

        # Only write this if we have something to write
        if flags:
            out.write(' ', attr.upper(), '="')
            # Omit the final space.
            out.write(*flags[:-1])
            out.write('"')

    def end(self):
        """Output package repository list."""
        out = self.out
        if self.verbosity > 0:
            super().end()
            out.write()
            if self.quiet_repo_display:
                repos = list(self.repos.items())
                repos.sort(key=operator.itemgetter(1))
                for k, v in repos:
                    reponame = getattr(k, 'repo_id', 'unknown repo id')
                    location = getattr(k, 'location', 'unspecified location')
                    if reponame != location:
                        self.out.write(
                            ' ', self.out.fg('cyan'), f"[{v}]",
                            self.out.reset, f" {reponame} ({location})")
                    else:
                        self.out.write(
                            ' ', self.out.fg('cyan'), f"[{v}]",
                            self.out.reset, f" {location}")


class PortageVerboseFormatter(VerboseFormatter, PortageFormatter):
    """Formatter designed to resemble portage output in verbose mode."""



from pkgcore.resolver.pigeonholes import PigeonHoledSlots
from pkgcore.restrictions import restriction
from snakeoil.test import TestCase

from .test_choice_point import fake_package


class fake_blocker(restriction.base):

    __slots__ = ('key', 'blocks')

    def __init__(self, key, blocks=()):
        restriction.base.__init__(self)
        if not isinstance(blocks, (list, tuple)):
            blocks = [blocks]
        object.__setattr__(self, "key", key)
        object.__setattr__(self, "blocks", blocks)

    def __str__(self):
        return f"fake_atom({self.key}, {self.blocks})"

    def match(self, obj):
        for x in self.blocks:
            if x is obj:
                return True
        return False

class SlotTesting(TestCase):

    def test_add(self):
        c = PigeonHoledSlots()
        o = fake_package()
        self.assertEqual([], c.fill_slotting(o))
        # test that it doesn't invalidly block o when (innefficiently)
        # doing a re-add
        self.assertEqual([o], c.fill_slotting(o))
        self.assertEqual([o], c.fill_slotting(fake_package()))
        self.assertFalse(c.fill_slotting(fake_package(slot=1, key=1)))

    def test_add_limiter(self):
        c = PigeonHoledSlots()
        p = fake_package()
        o = fake_blocker(None, p)
        self.assertFalse([], c.fill_slotting(p))
        self.assertEqual([p], c.add_limiter(o))
        c.remove_slotting(p)
        self.assertEqual([o], c.fill_slotting(p))
        # note we're doing 'is' tests in fake_blocker
        self.assertFalse([], c.fill_slotting(fake_package()))

    def test_remove_slotting(self):
        c = PigeonHoledSlots()
        p, p2 = fake_package(), fake_package(slot=2)
        o = fake_blocker(None, p)
        self.assertFalse([], c.add_limiter(o))
        c.remove_limiter(o)
        self.assertRaises(KeyError, c.remove_slotting, o)
        self.assertRaises(KeyError, c.remove_limiter, o)
        self.assertFalse([], c.fill_slotting(p))
        self.assertFalse([], c.fill_slotting(p2))
        c.remove_slotting(p)
        c.remove_slotting(p2)



"""
per key file based backend
"""

__all__ = ("database",)

import os
import stat

from snakeoil.fileutils import readlines_utf8
from snakeoil.osutils import pjoin

from ..config.hint import ConfigHint
from . import errors, fs_template


class database(fs_template.FsBased):
    """Stores cache entries in key=value form, stripping newlines."""

    # TODO: different way of passing in default auxdbkeys and location
    pkgcore_config_type = ConfigHint(
        {'readonly': 'bool', 'location': 'str', 'label': 'str',
         'auxdbkeys': 'list'},
        required=['location'],
        positional=['location'],
        typename='cache')

    autocommits = True
    mtime_in_entry = True
    eclass_chf_types = ('eclassdir', 'mtime')

    def _getitem(self, cpv):
        path = pjoin(self.location, cpv)
        try:
            data = readlines_utf8(path, True, True, True)
            if data is None:
                raise KeyError(cpv)
            return self._parse_data(data, data.mtime)
        except (EnvironmentError, ValueError) as e:
            raise errors.CacheCorruption(cpv, e) from e

    def _parse_data(self, data, mtime):
        d = self._cdict_kls()
        known = self._known_keys
        for x in data:
            k, v = x.split("=", 1)
            if k in known:
                d[k] = v

        if self._mtime_used:
            if self.mtime_in_entry:
                d[self._chf_key] = self._chf_deserializer(d[self._chf_key])
            else:
                d[self._chf_key] = int(mtime)
        else:
            d[self._chf_key] = self._chf_deserializer(d[self._chf_key])
        return d

    def _setitem(self, cpv, values):
        # might seem weird, but we rely on the trailing +1; this
        # makes it behave properly for any cache depth (including no depth)
        s = cpv.rfind('/') + 1
        fp = pjoin(self.location, cpv[:s], f'.update.{os.getpid()}.{cpv[s:]}')
        try:
            myf = open(fp, "w", 32768)
        except FileNotFoundError:
            if not self._ensure_dirs(cpv):
                raise errors.CacheCorruption(
                    cpv, f'error creating directory for {fp!r}')
            try:
                myf = open(fp, "w", 32768)
            except EnvironmentError as e:
                raise errors.CacheCorruption(cpv, e) from e
        except OSError as e:
            raise errors.CacheCorruption(cpv, e) from e

        if self._mtime_used:
            if not self.mtime_in_entry:
                mtime = values['_mtime_']
        for k, v in sorted(values.items()):
            myf.writelines(f'{k}={v}\n')

        myf.close()
        if self._mtime_used and not self.mtime_in_entry:
            self._ensure_access(fp, mtime=mtime)
        else:
            self._ensure_access(fp)

        # update written, now we move it
        new_fp = pjoin(self.location, cpv)
        try:
            os.rename(fp, new_fp)
        except EnvironmentError as e:
            os.remove(fp)
            raise errors.CacheCorruption(cpv, e) from e

    def _delitem(self, cpv):
        try:
            os.remove(pjoin(self.location, cpv))
        except FileNotFoundError:
            raise KeyError(cpv)
        except OSError as e:
            raise errors.CacheCorruption(cpv, e) from e

    def __contains__(self, cpv):
        return os.path.exists(pjoin(self.location, cpv))

    def keys(self):
        """generator for walking the dir struct"""
        dirs = [self.location]
        len_base = len(self.location)
        # Note: the misc try/except clauses are to protect against concurrent
        # modification of the cache resulting in transient errors.
        while dirs:
            d = dirs.pop(0)
            try:
                subdirs = os.listdir(d)
            except FileNotFoundError:
                continue
            except EnvironmentError as e:
                raise KeyError(cpv, f"access failure: {e}")
            for l in os.listdir(d):
                if l.endswith(".cpickle"):
                    continue
                p = pjoin(d, l)
                try:
                    st = os.lstat(p)
                except FileNotFoundError:
                    continue
                except EnvironmentError as e:
                    raise KeyError(cpv, f"Unhandled IO error: {e}")
                if stat.S_ISDIR(st.st_mode):
                    dirs.append(p)
                    continue
                yield p[len_base+1:]


class md5_cache(database):

    chf_type = 'md5'
    eclass_chf_types = ('md5',)
    chf_base = 16

    def __init__(self, location, **config):
        location = pjoin(location, 'metadata', 'md5-cache')
        super().__init__(location, **config)



from pkgcore.ebuild import conditionals
from pkgcore.ebuild.atom import atom
from pkgcore.ebuild.errors import DepsetParseError
from pkgcore.restrictions import boolean, packages
from snakeoil.currying import post_curry
from snakeoil.iterables import expandable_chain
from snakeoil.sequences import iflatten_instance
from snakeoil.test import TestCase


class base(TestCase):

    class kls(conditionals.DepSet):
        __slots__ = ()
        parse_depset = None

    def gen_depset(self, string, operators=None, element_kls=str,
                   element_func=None, **kwds):
        if element_func is not None:
            kwds["element_func"] = element_func
        if operators is None:
            operators = {"":boolean.AndRestriction, "||":boolean.OrRestriction}
        return self.kls.parse(string, element_kls, operators=operators, **kwds)


class TestDepSetParsing(base):

    def f(self, x):
        self.assertRaises(DepsetParseError, self.gen_depset, x)

    # generate a lot of parse error assertions.
    for idx, x in enumerate(("( )", "( a b c", "(a b c )",
        "( a b c)", "()", "x?( a )",
        "?x (a)", "x? (a )", "x? (a)", "x? ( a b)",
        "x? ( x? () )", "x? ( x? (a)", "(", ")", "x?",
        "||(", "||()", "||( )", "|| ()",
        "|| (", "|| )", "||)",  "|| ( x? ( )",
        "|| ( x?() )", "|| (x )", "|| ( x)",
        "a|", "a?", "a(b", "a)", "a||b",
        "a(", "a)b", "x? y", "( x )?", "||?")):
        locals()["test_DepsetParseError_case%i" % (idx + 1)] = post_curry(f, x)
    del x

    @staticmethod
    def mangle_cond_payload(p):
        l = [p]
        if isinstance(p, boolean.AndRestriction):
            l = iter(p)
        for x in l:
            s = ""
            if x.negate:
                s = "!"
            for y in x.vals:
                yield s + y

    def flatten_restricts(self, v):
        i = expandable_chain(v)
        depth = 0
        conditionals = []
        for x in i:
            for t, s in ((boolean.OrRestriction, "||"),
                         (boolean.AndRestriction, "&&")):
                if isinstance(x, t):
                    yield s
                    yield "("
                    i.appendleft(")")
                    i.appendleft(x.restrictions)
                    depth += 1
                    break
            else:
                if isinstance(x, packages.Conditional):
                    self.assertTrue(x.attr == "use")
                    conditionals.insert(
                        depth, list(self.mangle_cond_payload(x.restriction)))
                    yield set(iflatten_instance(conditionals[:depth + 1]))
                    yield "("
                    i.appendleft(")")
                    i.appendleft(x.payload)
                    depth += 1
                else:
                    if x == ")":
                        self.assertTrue(depth)
                        depth -= 1
                    yield x
        self.assertFalse(depth)

    def check_depset(self, s, func=base.gen_depset):
        if isinstance(s, (list, tuple)):
            s, v = s
            v2 = []
            for idx, x in enumerate(v):
                if isinstance(x, (list, tuple)):
                    v2.append(set(x))
                else:
                    v2.append(x)
            v = v2
        else:
            v = s.split()
        got = list(self.flatten_restricts(func(self, s)))
        wanted = list(v)
        self.assertEqual(got, v, msg="given {s!r}\nexpected {wanted!r} but got {got!r}")

    def check_str(self, s, func=base.gen_depset):
        if isinstance(s, (list, tuple)):
            s, v = s
            v2 = []
            for x in v:
                if isinstance(x, str):
                    v2.append(x)
                else:
                    v2.append(x[-1] + '?')
            v = ' '.join(v2)
        else:
            v = ' '.join(s.split())
        v = ' '.join(v.replace("&&", "").split())
        self.assertEqual(str(func(self, s)), v)

    # generate a lot of assertions of parse results.
    # if it's a list, first arg is string, second is results, if
    # string, the results for testing are determined by splitting the string
    for idx, x in enumerate([
        "a b",
        ( "",   []),

        ( "( a b )", ("&&", "(", "a", "b", ")")),

        "|| ( a b )",

        ( "a || ( a ( b  ) c || ( d )  )",
            ["a", "||", "(", "a", "b", "c", "d", ")"]),

        ( " x? ( a  b )",
            (["x"], "(", "a", "b", ")")),

        ( "x? ( y? ( a ) )",
            (["x"], "(", ["x", "y"], "(", "a", ")", ")")),

        ("|| ( || ( a b ) )", ["||", "(", "a", "b", ")"]),

        "|| ( || ( a b ) c )",

        ( "x? ( a !y? ( || ( b c ) d ) e ) f1 f? ( g h ) i",
            (
            ["x"], "(", "a", ["x", "!y"], "(", "||", "(", "b",
            "c", ")", "d", ")", "e", ")", "f1",
            ["f"], "(", "g", "h", ")", "i"
            )
        )]):

        locals()["test_parse_case%i" % (idx + 1)] = post_curry(check_depset, x)
        locals()["test_str_case%i" % (idx + 1)] = post_curry(check_str, x)

    def check_known_conditionals(self, text, conditionals, **kwds):
        d = self.gen_depset(text, **kwds)
        self.assertEqual(sorted(d.known_conditionals),
            sorted(conditionals.split()))
        # ensure it does the lookup *once*
        object.__setattr__(d, 'restrictions', ())
        self.assertFalse(d.restrictions)
        self.assertEqual(sorted(d.known_conditionals),
            sorted(conditionals.split()))

    for idx, (x, c) in enumerate([
        ["a? ( b )", "a"],
        ["a? ( b a? ( c ) )", "a"],
        ["a b c d e ( f )", ""],
        ["!a? ( b? ( c ) )", "a b"]
        ]):
        locals()["test_known_conditionals_case%i" % (idx + 1)] = post_curry(
            check_known_conditionals, x, c)
    del x, c

    def test_known_conditionals_transitive_use(self):
        self.check_known_conditionals(
            "a/b[c=] a/b[!d=] b/a[e?] b/a[!f?]", "c d e f", element_func=atom,
                transitive_use_atoms=True)

        self.check_known_conditionals(
            "|| ( b/a[e?] a/c )", "e", element_func=atom,
                transitive_use_atoms=True)

    def test_element_func(self):
        self.assertEqual(
            self.gen_depset("asdf fdas", element_func=post_curry(str)).element_class,
            "".__class__)

    def test_disabling_or(self):
        self.assertRaises(
            DepsetParseError, self.gen_depset, "|| ( a b )",
            {"operators":{"":boolean.AndRestriction}})

    def test_atom_interaction(self):
        self.gen_depset("a/b[x(+)]", element_func=atom)


class TestDepSetConditionalsInspection(base):

    def test_sanity_has_conditionals(self):
        self.assertFalse(bool(self.gen_depset("a b").has_conditionals))
        self.assertFalse(bool(
                self.gen_depset("( a b ) || ( c d )").has_conditionals))
        self.assertTrue(bool(self.gen_depset("x? ( a )").has_conditionals))
        self.assertTrue(bool(self.gen_depset("( x? ( a ) )").has_conditionals))
        self.assertTrue(bool(self.gen_depset("|| ( a/b[c=] b/d )", element_kls=atom,
            transitive_use_atoms=True).has_conditionals))

    def flatten_cond(self, c):
        l = set()
        for x in c:
            if isinstance(x, boolean.base):
                self.assertEqual(len(x.dnf_solutions()), 1)
                f = x.dnf_solutions()[0]
            else:
                f = [x]
            t = set()
            for a in f:
                s = ""
                if a.negate:
                    s = "!"
                t.update([f"{s}{y}" for y in a.vals])
            l.add(frozenset(t))
        return l

    def check_conds(self, s, r, msg=None, element_kls=str, **kwds):
        nc = {k: self.flatten_cond(v) for k, v in
              self.gen_depset(s, element_kls=element_kls, **kwds).node_conds.items()}
        d = {element_kls(k): v for k, v in r.items()}
        for k, v in d.items():
            if isinstance(v, str):
                d[k] = set([frozenset(v.split())])
            elif isinstance(v, (tuple, list)):
                d[k] = set(map(frozenset, v))

        self.assertEqual(nc, d, msg)

    for idx, s in enumerate((
        ("x? ( y )", {"y":"x"}),
        ("x? ( y ) z? ( y )", {"y":["z", "x"]}),
        ("x? ( z? ( w? ( y ) ) )", {"y":"w z x"}),
        ("!x? ( y )", {"y":"!x"}),
        ("!x? ( z? ( y a ) )", {"y":"!x z", "a":"!x z"}),
        ("x ( y )", {}),
        ("x ( y? ( z ) )", {"z":"y"}, "needs to dig down as deep as required"),
        ("x y? ( x )", {}, "x isn't controlled by a conditional, shouldn't be "
         "in the list"),
        ("|| ( y? ( x ) x )", {}, "x cannot be filtered down since x is "
         "accessible via non conditional path"),
        ("|| ( y? ( x ) z )", {"x":"y"}),
        )):
        locals()["test_node_conds_case%i" % (idx + 1)] = post_curry(check_conds, *s)

    for idx, s in enumerate((
        ("a/b[c=]", {"a/b[c]":"c", "a/b[-c]":"!c"}),
        )):
        locals()["test_node_conds_atom_%i" % (idx + 1)] = post_curry(check_conds,
            element_kls=atom, transitive_use_atoms=True, *s)


class TestDepSetEvaluate(base):

    def test_evaluation(self):
        flag_set = list(sorted("x%i" % (x,) for x in range(2000)))
        for vals in (
            ("y", "x? ( y ) !x? ( z )", "x"),
            ("z", "x? ( y ) !x? ( z )"),
            ("", "x? ( y ) y? ( z )"),
            ("a b", "a !x? ( b )"),
            ("a b", "a !x? ( b )", "", ""),
            ("a b", "a !x? ( b ) y? ( c )", "", "y"),
            ("a || ( b c )", "a || ( x? ( b ) c )", "x"),
            ("a c", "a || ( x? ( b ) c )"),
            ("a", "a x? ( y? ( b ) )"),
            ("a b", "a b"),
            ("a/b[-c]", "a/b[c=]"),
            ("a/b[c]", "a/b[c=]", "c"),
            ("a/b", "a/b[c?]"),
            ("a/b[-c]", "a/b[!c?]"),
            ("a/b", "a/b[!c?]", "c"),
            ("a/b[c]", "a/b[c?]", "c"),
            # this needs to a *very* large number of attributes; what we're asserting here
            # is that the backend doesn't use a quadratic implementation- if it does,
            # we want to preferably blow the allowed stack depth (runtime exception, but that's fine),
            # worst case (jython), we want to force a memory exhaustion.
            # we assert it in the tests to make sure some 'special' ebuild dev doesn't trigger
            # it on a user's machine, thus the abuse leveled here.
            ("a/b", "a/b[!c?,%s]" % (",".join(x + "?" for x in flag_set)), "c"),
            ("a/b", "a/b[%s]" % (",".join("%s?" % (x,) for x in flag_set)), "",
                " ".join(flag_set)),
            ("a/b[c,x0]", "a/b[c?,%s]" % (",".join(x + "?" for x in flag_set)), "c",
                " ".join(flag_set[1:])),
            ("a/b[c,%s]" % (','.join(flag_set),),
                "a/b[c?,%s]" % (",".join(x + "?" for x in flag_set)), "c",
                ""),
            ):

            result = vals[0]
            src = vals[1]
            use, tristate, kls = [], None, str
            if len(vals) > 2:
                use = vals[2].split()
            if len(vals) > 3:
                tristate = vals[3].split()
            kwds = {}
            if '/' in src:
                kls = atom
                flags = src.split("[", 1)[-1]
                if "?" in flags or "=" in flags:
                    kwds['transitive_use_atoms'] = True
            else:
                kls = str
            orig = self.gen_depset(src, element_kls=kls, **kwds)
            collapsed = orig.evaluate_depset(use,
                tristate_filter=tristate)
            self.assertEqual(str(collapsed), result, msg=
                "expected %r got %r\nraw depset: %r\nuse: %r, tristate: %r" %
                    (result, str(collapsed), src, use, tristate))
            if not ('?' in src or kwds.get("transitive_use_atoms")):
                self.assertIdentical(orig, collapsed)



"""Utilities for writing commandline utilities.

pkgcore scripts should use the :obj:`ArgumentParser` subclass here for a
consistent commandline "look and feel" (and it tries to make life a
bit easier too). They will probably want to use :obj:`main` from an C{if
__name__ == '__main__'} block too: it will take care of things like
consistent exception handling.

See dev-notes/commandline.rst for more complete documentation.
"""

import argparse
import os
import sys
from functools import partial
from importlib import import_module

from snakeoil import modules
from snakeoil.cli import arghparse, tool
from snakeoil.log import suppress_logging
from snakeoil.osutils import abspath, normpath, pjoin
from snakeoil.sequences import iflatten_instance, unstable_unique
from snakeoil.strings import pluralism

from .. import const
from ..config import load_config
from ..plugin import get_plugins
from ..repository import errors as repo_errors
from ..restrictions import packages, restriction
from . import parserestrict


class StoreTarget(argparse._AppendAction):
    """Parse extended package atom syntax and optionally set arguments.

    Various target arguments are supported including the following:

    atom
        An extended atom syntax is supported, see the related section
        in pkgcore(5).

    package set
        Used to define lists of packages, the syntax used for these is
        @pkgset. For example, the @system and @world package sets are
        supported.

    extended globbing
        Globbing package names or atoms allows for use cases such as
        ``'far*'`` (merge every package starting with 'far'),
        ``'dev-python/*::gentoo'`` (merge every package in the dev-python
        category from the gentoo repo), or even '*' (merge everything).

    Also, the target '-' allows targets to be read from standard input.
    """

    def __init__(self, *args, **kwargs):
        self.use_sets = kwargs.pop('use_sets', False)
        self.allow_ebuild_paths = kwargs.pop('allow_ebuild_paths', False)
        self.allow_external_repos = kwargs.pop('allow_external_repos', False)
        self.separator = kwargs.pop('separator', None)
        kwargs.setdefault('default', ())
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if self.separator is not None:
            values = values.split(self.separator)
        if self.use_sets:
            setattr(namespace, self.use_sets, [])

        if isinstance(values, str):
            values = [values]
        elif values is not None and len(values) == 1 and values[0] == '-':
            if not sys.stdin.isatty():
                values = [x.strip() for x in sys.stdin.readlines() if x.strip() != '']
                # reassign stdin to allow interactivity (currently only works for unix)
                sys.stdin = open('/dev/tty')
            else:
                raise argparse.ArgumentError(self, "'-' is only valid when piping data in")

        # override default empty tuple value to appendable list
        if values:
            setattr(namespace, self.dest, [])

        for token in values:
            if self.use_sets and token.startswith('@'):
                namespace.sets.append(token[1:])
            else:
                if self.allow_ebuild_paths and token.endswith('.ebuild'):
                    try:
                        repo = getattr(namespace, 'repo', namespace.domain.ebuild_repos_raw)
                    except AttributeError:
                        raise argparse.ArgumentTypeError(
                            'repo or domain must be defined in the namespace')
                    if not os.path.exists(token):
                        raise argparse.ArgumentError(self, f"nonexistent ebuild: {token!r}")
                    elif not os.path.isfile(token):
                        raise argparse.ArgumentError(self, f"invalid ebuild: {token!r}")
                    if self.allow_external_repos and token not in repo:
                        repo_root_dir = os.path.abspath(
                            pjoin(token, os.pardir, os.pardir, os.pardir))
                        try:
                            with suppress_logging():
                                repo = namespace.domain.add_repo(
                                    repo_root_dir, config=namespace.config)
                        except repo_errors.RepoError as e:
                            raise argparse.ArgumentError(self, f"{token!r} -- {e}")
                    try:
                        restriction = repo.path_restrict(token)
                    except ValueError as e:
                        raise argparse.ArgumentError(self, e)
                else:
                    try:
                        restriction = parserestrict.parse_match(token)
                    except parserestrict.ParseError as e:
                        parser.error(e)
                super().__call__(
                    parser, namespace,
                    (token, restriction), option_string=option_string)


CONFIG_ALL_DEFAULT = object()


class NoDefaultConfigError(argparse.ArgumentError):
    pass


class StoreConfigObject(argparse._StoreAction):

    default_priority = 20

    def __init__(self, *args, **kwargs):
        self.priority = int(kwargs.pop("priority", self.default_priority))
        self.config_type = kwargs.pop("config_type", None)
        if self.config_type is None or not isinstance(self.config_type, str):
            raise ValueError("config_type must specified, and be a string")

        if kwargs.pop("get_default", False):
            kwargs["default"] = arghparse.DelayedValue(
                partial(self.store_default, self.config_type,
                        option_string=kwargs.get('option_strings', [None])[0]),
                self.priority)

        self.store_name = kwargs.pop("store_name", False)
        self.writable = kwargs.pop("writable", None)
        self.target = argparse._StoreAction(*args, **kwargs)

        super().__init__(*args, **kwargs)

    @staticmethod
    def _choices(sections):
        """Yield available values for a given option."""
        for k, v in sections.items():
            yield k

    def _load_obj(self, sections, name):
        obj_type = self.metavar if self.metavar is not None else self.config_type
        obj_type = obj_type.lower() + ' ' if obj_type is not None else ''

        try:
            val = sections[name]
        except KeyError:
            choices = ', '.join(self._choices(sections))
            if choices:
                choices = f" (available: {choices})"

            raise argparse.ArgumentError(
                self, f"couldn't find {obj_type}{name!r}{choices}")

        if self.writable and getattr(val, 'frozen', False):
            raise argparse.ArgumentError(
                self, f"{obj_type}{name!r} is readonly")

        if self.store_name:
            return name, val
        return val

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, arghparse.DelayedParse(
            partial(self._real_call, parser, namespace, values, option_string),
            self.priority))

    def _get_sections(self, config, namespace):
        return getattr(config, self.config_type)

    def _real_call(self, parser, namespace, values, option_string=None):
        config = getattr(namespace, 'config', None)
        if config is None:
            raise ValueError("no config found, internal bug")

        sections = self._get_sections(config, namespace)

        if self.nargs == argparse.ZERO_OR_MORE and values == []:
            values = list(sections.keys())

        if values is CONFIG_ALL_DEFAULT:
            value = [self._load_obj(sections, x) for x in sections]
        elif isinstance(values, str):
            value = self._load_obj(sections, values)
        else:
            value = [self._load_obj(sections, x) for x in values]
        setattr(namespace, self.dest, value)

    @staticmethod
    def store_default(config_type, namespace, attr, option_string=None):
        config = getattr(namespace, 'config', None)
        if config is None:
            raise argparse.ArgumentTypeError(
                "no config found -- internal bug, or broken on disk configuration")
        obj = config.get_default(config_type)
        if obj is None:
            known_objs = sorted(getattr(config, config_type).keys())
            msg = f"config error: no default object of type {config_type!r} found.  "
            if not option_string:
                msg += "Please fix your configuration."
            else:
                msg += (
                    "Please either fix your configuration, or set the "
                    f"{config_type} via the {option_string} option.")
            if known_objs:
                msg += f"Known {config_type}s: {', '.join(map(repr, known_objs))}"
            raise NoDefaultConfigError(None, msg)
        setattr(namespace, attr, obj)

    @staticmethod
    def store_all_default(config_type, namespace, attr):
        config = getattr(namespace, 'config', None)
        if config is None:
            raise ValueError("no config found -- internal bug")
        obj = [(k, v) for k, v in getattr(config, config_type).items()]
        setattr(namespace, attr, obj)

    @classmethod
    def lazy_load_object(cls, config_type, key, priority=None):
        if priority is None:
            priority = cls.default_priority
        return arghparse.DelayedValue(
            partial(cls._lazy_load_object, config_type, key),
            priority)

    @staticmethod
    def _lazy_load_object(config_type, key, namespace, attr):
        try:
            obj = getattr(namespace.config, config_type)[key]
        except KeyError:
            raise argparse.ArgumentError(
                None, f"couldn't find {config_type} {attr!r}")
        setattr(namespace, attr, obj)


class StoreRepoObject(StoreConfigObject):
    """Load a repo object from the config."""

    # mapping between supported repo type requests and the related attr on
    # domain objects to pull the requested repos from
    valid_repo_types = {
        'config': 'repo_configs',
        'all': 'repos',
        'all-raw': 'repos_raw',
        'source': 'source_repos',
        'source-raw': 'source_repos_raw',
        'installed': 'installed_repos',
        'installed-raw': 'installed_repos_raw',
        'unfiltered': 'unfiltered_repos',
        'ebuild': 'ebuild_repos',
        'ebuild-unfiltered': 'ebuild_repos_unfiltered',
        'ebuild-raw': 'ebuild_repos_raw',
        'binary': 'binary_repos',
        'binary-unfiltered': 'binary_repos_unfiltered',
        'binary-raw': 'binary_repos_raw',
    }

    def __init__(self, *args, **kwargs):
        if 'config_type' in kwargs:
            raise ValueError(
                "StoreRepoObject: config_type keyword is redundant: got %s"
                % (kwargs['config_type'],))

        self.repo_type = kwargs.pop('repo_type', 'all')
        if self.repo_type not in self.valid_repo_types:
            raise argparse.ArgumentTypeError(f"unknown repo type: {self.repo_type!r}")
        self.repo_key = self.valid_repo_types[self.repo_type]

        self.allow_aliases = set(kwargs.pop("allow_aliases", ()))
        if self.allow_aliases:
            unknown_aliases = self.allow_aliases.difference(self.valid_repo_types)
            if unknown_aliases:
                raise argparse.ArgumentTypeError(
                    'unknown repo alias%s: %s' % (
                        pluralism(unknown_aliases, plural='es'), ', '.join(unknown_aliases)))

        if self.repo_type == 'config':
            kwargs['config_type'] = 'repo_config'
        else:
            kwargs['config_type'] = 'repo'
        self.allow_name_lookup = kwargs.pop("allow_name_lookup", True)
        self.allow_external_repos = kwargs.pop("allow_external_repos", False)
        super().__init__(*args, **kwargs)

    def _get_sections(self, config, namespace):
        domain = getattr(namespace, 'domain', None)

        # return repo config objects
        if domain is None or self.repo_type == 'config':
            return StoreConfigObject._get_sections(self, config, namespace)

        self.config = config
        self.domain = config.get_default("domain")

        # return the type of repos requested
        return getattr(self.domain, self.repo_key)

    @staticmethod
    def _choices(sections):
        """Return an iterable of name: location mappings for available repos.

        If a repo doesn't have a proper location just the name is returned.
        """
        for repo_name, repo in sorted(unstable_unique(sections.items())):
            repo_name = getattr(repo, 'repo_id', repo_name)
            if hasattr(repo, 'location'):
                yield f"{repo_name}:{repo.location}"
            else:
                yield repo_name

    def _load_obj(self, sections, name):
        repo = name
        if not self.allow_name_lookup or repo in sections:
            # requested repo exists in the config
            pass
        elif name in self.allow_aliases and self.valid_repo_types[name]:
            # pull repos related to given alias
            return getattr(self.domain, self.valid_repo_types[name])
        else:
            # name wasn't found, check repo aliases for it
            for repo_name, repo_obj in sections.items():
                if repo in repo_obj.aliases:
                    repo = repo_name
                    break
            else:
                # try to add it as an external repo
                if self.allow_external_repos and os.path.exists(repo):
                    try:
                        configure = not self.repo_type.endswith('-raw')
                        with suppress_logging():
                            repo_obj = self.domain.add_repo(
                                repo, config=self.config, configure=configure)
                        repo = repo_obj.location
                    except repo_errors.RepoError as e:
                        raise argparse.ArgumentError(self, e)
                    if hasattr(self.domain, '_' + self.repo_key):
                        # force JIT-ed attr refresh to include newly added repo
                        setattr(self.domain, '_' + self.repo_key, None)
                    sections = getattr(self.domain, self.repo_key)
        return StoreConfigObject._load_obj(self, sections, repo)


class DomainFromPath(StoreConfigObject):

    def __init__(self, *args, **kwargs):
        kwargs['config_type'] = 'domain'
        super().__init__(*args, **kwargs)

    def _load_obj(self, sections, requested_path):
        targets = list(find_domains_from_path(sections, requested_path))
        if not targets:
            raise ValueError(f"couldn't find domain at path {requested_path!r}")
        elif len(targets) != 1:
            raise ValueError(
                "multiple domains claim root %r: domains %s" %
                (requested_path, ', '.join(repr(x[0]) for x in targets)))
        return targets[0][1]


def find_domains_from_path(sections, path):
    path = normpath(abspath(path))
    for name, domain in sections.items():
        root = getattr(domain, 'root', None)
        if root is None:
            continue
        root = normpath(abspath(root))
        if root == path:
            yield name, domain


class BooleanQuery(arghparse.DelayedValue):

    def __init__(self, attrs, klass_type=None, priority=100, converter=None):
        if klass_type == 'and':
            self.klass = packages.AndRestriction
        elif klass_type == 'or':
            self.klass = packages.OrRestriction
        elif callable(klass_type):
            self.klass = klass_type
        else:
            raise ValueError(
                "klass_type either needs to be 'or', 'and', "
                f"or a callable. Got {klass_type!r}")

        if converter is not None and not callable(converter):
            raise ValueError(
                "converter either needs to be None, or a callable;"
                f" got {converter!r}")

        self.converter = converter
        self.priority = int(priority)
        self.attrs = tuple(attrs)

    def invokable(self, namespace, attr):
        l = []
        for x in self.attrs:
            val = getattr(namespace, x, None)
            if val is None:
                continue
            if isinstance(val, bool):
                # Skip converter call for disabled boolean actions
                if not val:
                    self.converter = False
            elif isinstance(val, restriction.base):
                l.append(val)
            else:
                l.extend(val)

        if self.converter:
            l = self.converter(l, namespace)

        l = list(iflatten_instance(l, (restriction.base,)))

        if len(l) > 1:
            val = self.klass(*l)
        elif l:
            val = l[0]
        else:
            val = None
        setattr(namespace, attr, val)


def make_query(parser, *args, **kwargs):
    klass_type = kwargs.pop("klass_type", "or")
    dest = kwargs.pop("dest", None)
    if dest is None:
        raise TypeError("dest must be specified via kwargs")
    attrs = kwargs.pop("attrs", [])
    subattr = f"_{dest}"
    kwargs["dest"] = subattr
    if kwargs.get('type', False) is None:
        del kwargs['type']
    else:
        def query(value):
            return parserestrict.parse_match(value)
        kwargs.setdefault("type", query)
    if kwargs.get('metavar', False) is None:
        del kwargs['metavar']
    else:
        kwargs.setdefault("metavar", dest)
    final_priority = kwargs.pop("final_priority", None)
    final_converter = kwargs.pop("final_converter", None)
    parser.add_argument(*args, **kwargs)
    bool_kwargs = {'converter': final_converter}
    if final_priority is not None:
        bool_kwargs['priority'] = final_priority
    obj = BooleanQuery(list(attrs) + [subattr], klass_type=klass_type, **bool_kwargs)
    # note that dict expansion has to be used here; dest=obj would just set a
    # default named 'dest'
    parser.set_defaults(**{dest: obj})


def python_namespace_type(value, module=False, attribute=False):
    """
    return the object from python namespace that value specifies

    :param value: python namespace, snakeoil.modules for example
    :param module: if true, the object must be a module
    :param attribute: if true, the object must be a non-module
    :raises ValueError: if the conditions aren't met, or import fails
    """
    try:
        if module:
            return import_module(value)
        elif attribute:
            return modules.load_attribute(value)
        return modules.load_any(value)
    except (ImportError, modules.FailedImport) as err:
        raise argparse.ArgumentTypeError(str(err)) from err


def register_command(commands, real_type=type):
    def f(name, bases, scope, real_type=real_type, commands=commands):
        o = real_type(name, bases, scope)
        commands.append(o)
        return o
    return f


def store_config(namespace, attr, global_config=()):
    config = load_config(
        prepend_sources=tuple(global_config),
        location=namespace.config_path,
        **vars(namespace))
    setattr(namespace, attr, config)


def _mk_domain(parser, help=True):
    parser.add_argument(
        '--domain', get_default=True, config_type='domain',
        action=StoreConfigObject,
        help='custom pkgcore domain to use for this operation' if help else argparse.SUPPRESS)


class _SubParser(arghparse._SubParser):

    def add_parser(self, name, config=False, domain=False, **kwds):
        """Suppress config and domain options in subparsers by default.

        They are rarely used so only allow them as options to the base command.
        """
        return super().add_parser(name, config=config, domain=domain, **kwds)


class _ConfigArg(argparse._StoreAction):
    """Store given config path location or use the stub config when disabled."""

    def __call__(self, parser, namespace, value, option_string=None):
        if value.lower() in {'false', 'no', 'n'}:
            path = pjoin(const.DATA_PATH, 'stubconfig')
        else:
            path = arghparse.existent_path(value)
        setattr(namespace, self.dest, path)


class ArgumentParser(arghparse.ArgumentParser):

    def __init__(self, suppress=False, help=True, config=True,
                 domain=True, script=None, **kwds):
        super().__init__(suppress=suppress, script=script, **kwds)
        self.register('action', 'parsers', _SubParser)

        if not suppress:
            config_opts = self.add_argument_group('config options')
            if config:
                config_opts.add_argument(
                    '--config', action=_ConfigArg, dest='config_path',
                    help='use custom config or skip loading system config' if help else argparse.SUPPRESS,
                    docs="""
                        The path to a custom pkgcore config file or portage
                        config directory can be given to override loading the
                        default system config.

                        Alternatively, an argument of 'false' or 'no' will skip
                        loading the system config entirely if one exists.
                    """)

                if script is not None:
                    try:
                        _, script_module = script
                    except TypeError:
                        raise ValueError(
                            "invalid script parameter, should be (__file__, __name__)")
                    project = script_module.split('.')[0]
                else:
                    project = __name__.split('.')[0]

                # TODO: figure out a better method for plugin registry/loading
                kwargs = {}
                try:
                    plugins = import_module('.plugins', project)
                    kwargs['global_config'] = get_plugins('global_config', plugins)
                except ImportError:
                    # project doesn't bundle plugins
                    pass
                self.set_defaults(config=arghparse.DelayedValue(
                    partial(store_config, **kwargs)))

            if domain:
                _mk_domain(config_opts, help)


def convert_to_restrict(sequence, default=packages.AlwaysTrue):
    """Convert an iterable to a list of atoms, or return the default"""
    l = []
    try:
        for x in sequence:
            l.append(parserestrict.parse_match(x))
    except parserestrict.ParseError as e:
        raise argparse.ArgumentError(f"invalid atom: {x!r}: {e}") from e
    return l or [default]


class Tool(tool.Tool):
    """pkgcore-specific commandline utility functionality."""

    def pre_parse(self, *args, **kwargs):
        """Pass down pkgcore-specific settings to the bash side."""
        # pass down verbosity level to affect debug output
        if self.parser.debug:
            os.environ['PKGCORE_DEBUG'] = str(self.parser.verbosity)



import datetime
import os
from unittest import mock

import pytest
from pkgcore.sync import base, rsync
from pkgcore.sync.tar import tar_syncer
from snakeoil.process import CommandNotFound


def fake_ips(num):
    """Generate simple IPv4 addresses given the amount to create."""
    return [
        (None, None, None, None, ('.'.join(str(x) * 4), 0))
        for x in range(num)
    ]


@mock.patch('socket.getaddrinfo', return_value=fake_ips(3))
@mock.patch('snakeoil.process.spawn.spawn')
class TestRsyncSyncer:

    _syncer_class = rsync.rsync_syncer

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.repo_path = str(tmp_path / 'repo')
        with mock.patch('snakeoil.process.find_binary', return_value='rsync'):
            self.syncer = self._syncer_class(
                self.repo_path, "rsync://rsync.gentoo.org/gentoo-portage")

    @mock.patch('snakeoil.process.find_binary')
    def test_uri_parse_rsync_missing(self, find_binary, spawn, getaddrinfo):
        find_binary.side_effect = CommandNotFound('rsync')
        with pytest.raises(base.SyncError):
            self._syncer_class(self.repo_path, 'rsync://foon.com/dar')

    @mock.patch('snakeoil.process.find_binary')
    def test_uri_parse(self, find_binary, spawn, getaddrinfo):
        find_binary.side_effect = lambda x: x
        o = self._syncer_class(self.repo_path, 'rsync://dar/module')
        assert o.uri == 'rsync://dar/module/'
        assert o.rsh == None

        o = self._syncer_class(self.repo_path, 'rsync+/bin/sh://dar/module')
        assert o.uri == 'rsync://dar/module/'
        assert o.rsh == '/bin/sh'

    def test_successful_sync(self, spawn, getaddrinfo):
        spawn.return_value = 0
        assert self.syncer.sync()
        spawn.assert_called_once()

    def test_bad_syntax_sync(self, spawn, getaddrinfo):
        spawn.return_value = 1
        with pytest.raises(base.SyncError) as excinfo:
            assert self.syncer.sync()
        assert str(excinfo.value).startswith('rsync command syntax error:')
        spawn.assert_called_once()

    def test_failed_disk_space_sync(self, spawn, getaddrinfo):
        spawn.return_value = 11
        with pytest.raises(base.SyncError) as excinfo:
            assert self.syncer.sync()
        assert str(excinfo.value) == 'rsync ran out of disk space'
        spawn.assert_called_once()

    def test_retried_sync(self, spawn, getaddrinfo):
        spawn.return_value = 99
        with pytest.raises(base.SyncError) as excinfo:
            assert self.syncer.sync()
        assert str(excinfo.value) == 'all attempts failed'
        # rsync should retry every resolved IP related to the sync URI
        assert len(spawn.mock_calls) == 3

    def test_retried_sync_max_retries(self, spawn, getaddrinfo):
        spawn.return_value = 99
        # generate more IPs than retries
        getaddrinfo.return_value = fake_ips(self.syncer.retries + 1)
        with pytest.raises(base.SyncError) as excinfo:
            assert self.syncer.sync()
        assert str(excinfo.value) == 'all attempts failed'
        assert len(spawn.mock_calls) == self.syncer.retries

    def test_failed_dns_sync(self, spawn, getaddrinfo):
        getaddrinfo.side_effect = OSError()
        with pytest.raises(base.SyncError) as excinfo:
            assert self.syncer.sync()
        assert str(excinfo.value).startswith('DNS resolution failed')
        spawn.assert_not_called()


class TestRsyncTimestampSyncer(TestRsyncSyncer):

    _syncer_class = rsync.rsync_timestamp_syncer


@pytest.mark_network
class TestRsyncSyncerReal:

    def test_sync(self, tmp_path):
        # perform a tarball sync for initial week-old base
        path = tmp_path / 'repo'
        week_old = datetime.datetime.now() - datetime.timedelta(days=7)
        date_str = week_old.strftime("%Y%m%d")
        syncer = tar_syncer(
            str(path), f"http://distfiles.gentoo.org/snapshots/portage-{date_str}.tar.xz")
        assert syncer.sync()
        timestamp = os.path.join(path, 'metadata', 'timestamp.chk')
        assert os.path.exists(timestamp)
        stat = os.stat(timestamp)

        # run rsync over the unpacked repo tarball to update to the latest tree
        syncer = rsync.rsync_timestamp_syncer(
            str(path), "rsync://rsync.gentoo.org/gentoo-portage")
        assert syncer.sync()
        assert stat != os.stat(timestamp)



__all__ = (
    "ProfileError", "ProfileNode", "EmptyRootNode", "OnDiskProfile",
    "UserProfile",
)

import os
from collections import defaultdict, namedtuple
from functools import partial
from itertools import chain

from snakeoil import caching, klass
from snakeoil.bash import iter_read_bash, read_bash_dict
from snakeoil.data_source import local_source
from snakeoil.fileutils import readlines_utf8
from snakeoil.mappings import ImmutableDict
from snakeoil.osutils import abspath, pjoin
from snakeoil.sequences import split_negations, stable_unique

from ..config import errors
from ..config.hint import ConfigHint
from ..fs.livefs import sorted_scan
from ..log import logger
from . import const, cpv
from . import errors as ebuild_errors
from . import misc, repo_objs
from .atom import atom
from .eapi import EAPI, get_eapi


class ProfileError(errors.ParsingError):

    def __init__(self, path, filename, error):
        self.path, self.filename, self.error = path, filename, error

    def __str__(self):
        if self.filename:
            return f"failed parsing {self.filename!r} in {self.path!r}: {self.error}"
        return f"failed parsing {self.path!r}: {self.error}"


class NonexistentProfile(ProfileError):
    """Profile for a nonexistent directory."""

    def __init__(self, path):
        super().__init__(path, "", "nonexistent profile directory")


def _read_profile_files(files, allow_line_cont=False):
    """Read all the given data files."""
    for path in files:
        # determine file path relative to the profiles dir
        try:
            relpath = path.split('/profiles/')[1]
        except IndexError:
            # profiles base path
            relpath = os.path.basename(path)

        for lineno, line in iter_read_bash(
                path, allow_line_cont=allow_line_cont, enum_line=True):
            yield line, lineno, relpath


def load_property(filename, *, read_func=_read_profile_files, fallback=(),
                  parse_func=lambda x: x, allow_line_cont=False, allow_recurse=False,
                  eapi_optional=None):
    """Decorator simplifying parsing profile files to generate a profile property.

    :param filename: The filename to parse within that profile directory.
    :keyword read_func: An invokable used to read the specified file.
    :keyword fallback: What to return if the file does not exist for this profile. Must be immutable.
    :keyword parse_func: An invokable used to parse the data.
    :keyword allow_line_cont: Controls whether line continuations are respected.
    :keyword allow_recurse: Controls whether or not this specific content can be a directory
        of files, rather than just a file.  Only is consulted if we're parsing the profile
        in non pms strict mode.
    :keyword eapi_optional: If given, the EAPI for this profile node is checked to see if
        the given optional evaluates to True; if so, then parsing occurs.  If False, then
        the fallback is returned and no ondisk activity occurs.
    :return: A :py:`klass.jit.attr_named` property instance.
    """
    def f(func):
        f2 = klass.jit_attr_named(f'_{func.__name__}')
        return f2(partial(
            _load_and_invoke, func, filename, read_func, fallback,
            allow_recurse, allow_line_cont, parse_func, eapi_optional))
    return f


def _load_and_invoke(func, filename, read_func, fallback, allow_recurse,
                     allow_line_cont, parse_func, eapi_optional, self):
    if eapi_optional is not None and not getattr(self.eapi.options, eapi_optional, None):
        return func(self, fallback)

    profile_path = self.path.rstrip('/')
    base = pjoin(profile_path, filename)

    files = []
    if self.pms_strict or not allow_recurse:
        if os.path.exists(base):
            files.append(base)
    else:
        # Skip hidden files and backup files, those beginning with '.' or
        # ending with '~', respectively.
        files.extend(sorted_scan(base, hidden=False, backup=False))

    try:
        if files:
            if read_func is None:
                data = parse_func(files)
            else:
                data = parse_func(read_func(
                    files, allow_line_cont=allow_line_cont))
        else:
            data = fallback
        return func(self, data)
    except (ValueError, IndexError, EnvironmentError) as e:
        raise ProfileError(profile_path, filename, e) from e
    except IsADirectoryError as e:
        raise ProfileError(
            self.path, filename,
            "path is a directory, but this profile is PMS format- "
            "directories aren't allowed. See layout.conf profile-formats "
            "to enable directory support") from e


_make_incrementals_dict = partial(misc.IncrementalsDict, const.incrementals)
_Packages = namedtuple("_Packages", ("system", "profile"))


class ProfileNode(metaclass=caching.WeakInstMeta):

    __inst_caching__ = True
    _repo_map = None

    def __init__(self, path, pms_strict=True):
        self.path = path.rstrip(os.path.sep)
        if not os.path.isdir(self.path):
            raise NonexistentProfile(self.path)
        self.pms_strict = pms_strict

    def __str__(self):
        return f"profile at {self.path!r}"

    def __repr__(self):
        return '<%s path=%r, @%#8x>' % (self.__class__.__name__, self.path, id(self))

    system = klass.alias_attr("packages.system")
    profile_set = klass.alias_attr("packages.profile")

    @klass.jit_attr
    def name(self):
        """Relative path to the profile from the profiles directory."""
        try:
            return self.path.split('/profiles/')[1]
        except IndexError:
            # profiles base path
            return ''

    @load_property("packages")
    def packages(self, data):
        repo_config = self.repoconfig
        # TODO: get profile-set support into PMS
        profile_set = repo_config is not None and 'profile-set' in repo_config.profile_formats
        sys, neg_sys, pro, neg_pro = [], [], [], []
        neg_wildcard = False
        for line, lineno, relpath in data:
            try:
                if line[0] == '-':
                    if line == '-*':
                        neg_wildcard = True
                    elif line[1] == '*':
                        neg_sys.append(self.eapi_atom(line[2:]))
                    elif profile_set:
                        neg_pro.append(self.eapi_atom(line[1:]))
                    else:
                        logger.error(f'{relpath!r}: invalid line format, line {lineno}: {line!r}')
                else:
                    if line[0] == '*':
                        sys.append(self.eapi_atom(line[1:]))
                    elif profile_set:
                        pro.append(self.eapi_atom(line))
                    else:
                        logger.error(f'{relpath!r}: invalid line format, line {lineno}: {line!r}')
            except ebuild_errors.MalformedAtom as e:
                logger.error(f'{relpath!r}, line {lineno}: parsing error: {e}')
        system = [tuple(neg_sys), tuple(sys)]
        profile = [tuple(neg_pro), tuple(pro)]
        if neg_wildcard:
            system.append(neg_wildcard)
            profile.append(neg_wildcard)
        return _Packages(tuple(system), tuple(profile))

    @load_property("parent")
    def parent_paths(self, data):
        repo_config = self.repoconfig
        if repo_config is not None and 'portage-2' in repo_config.profile_formats:
            l = []
            for line, lineno, relpath in data:
                repo_id, separator, profile_path = line.partition(':')
                if separator:
                    if repo_id:
                        try:
                            location = self._repo_map[repo_id]
                        except KeyError:
                            # check if requested repo ID matches the current
                            # repo which could be the case when running against
                            # unconfigured, external repos.
                            if repo_id == repo_config.repo_id:
                                location = repo_config.location
                            else:
                                logger.error(
                                    f'repo {repo_config.repo_id!r}: '
                                    f"{relpath!r} (line {lineno}), "
                                    f'bad profile parent {line!r}: '
                                    f'unknown repo {repo_id!r}'
                                )
                                continue
                    l.append((abspath(pjoin(location, 'profiles', profile_path)), line, lineno))
                else:
                    l.append((abspath(pjoin(self.path, repo_id)), line, lineno))
            return tuple(l)
        return tuple((abspath(pjoin(self.path, line)), line, lineno)
                     for line, lineno, relpath in data)

    @klass.jit_attr
    def parents(self):
        kls = getattr(self, 'parent_node_kls', self.__class__)
        parents = []
        for path, line, lineno in self.parent_paths:
            try:
                parents.append(kls(path))
            except ProfileError as e:
                repo_id = self.repoconfig.repo_id
                logger.error(
                    f"repo {repo_id!r}: '{self.name}/parent' (line {lineno}), "
                    f'bad profile parent {line!r}: {e.error}'
                )
                continue
        return tuple(parents)

    @load_property("package.provided", allow_recurse=True,
                   eapi_optional='profile_pkg_provided')
    def pkg_provided(self, data):
        def _parse_cpv(s):
            try:
                return cpv.VersionedCPV(s)
            except cpv.InvalidCPV:
                logger.error(f'invalid package.provided entry: {s!r}')
        data = (x[0] for x in data)
        return split_negations(data, _parse_cpv)

    def _parse_atom_negations(self, data):
        """Parse files containing optionally negated package atoms."""
        neg, pos = [], []
        for line, lineno, relpath in data:
            if line[0] == '-':
                line = line[1:]
                if not line:
                    logger.error(f"{relpath!r}, line {lineno}: '-' negation without an atom")
                    continue
                l = neg
            else:
                l = pos
            try:
                l.append(self.eapi_atom(line))
            except ebuild_errors.MalformedAtom as e:
                logger.error(f'{relpath!r}, line {lineno}: parsing error: {e}')
        return tuple(neg), tuple(pos)

    def _package_keywords_splitter(self, iterable):
        """Parse package keywords files."""
        for line, lineno, relpath in iterable:
            v = line.split()
            try:
                yield (atom(v[0]), tuple(stable_unique(v[1:])))
            except ebuild_errors.MalformedAtom as e:
                logger.error(f'{relpath!r}, line {lineno}: parsing error: {e}')

    @load_property("package.mask", allow_recurse=True)
    def masks(self, data):
        return self._parse_atom_negations(data)

    @load_property("package.unmask", allow_recurse=True)
    def unmasks(self, data):
        return self._parse_atom_negations(data)

    @load_property("package.deprecated", allow_recurse=True)
    def pkg_deprecated(self, data):
        return self._parse_atom_negations(data)

    @load_property("package.keywords", allow_recurse=True)
    def keywords(self, data):
        return tuple(self._package_keywords_splitter(data))

    @load_property("package.accept_keywords", allow_recurse=True)
    def accept_keywords(self, data):
        return tuple(self._package_keywords_splitter(data))

    @load_property("package.use", allow_recurse=True)
    def pkg_use(self, data):
        c = misc.ChunkedDataDict()
        c.update_from_stream(
            chain.from_iterable(self._parse_package_use(data).values()))
        c.freeze()
        return c

    @load_property("deprecated", read_func=None, fallback=None)
    def deprecated(self, data):
        if data is not None:
            data = iter(readlines_utf8(data[0]))
            try:
                replacement = next(data).strip()
                msg = "\n".join(x.lstrip("#").strip() for x in data)
                data = (replacement, msg)
            except StopIteration:
                data = None
        return data

    def _parse_package_use(self, data):
        d = defaultdict(list)
        # split the data down ordered cat/pkg lines
        for line, lineno, relpath in data:
            l = line.split()
            try:
                a = self.eapi_atom(l[0])
            except ebuild_errors.MalformedAtom as e:
                logger.error(f'{relpath!r}, line {lineno}: parsing error: {e}')
                continue
            if len(l) == 1:
                logger.error(f'{relpath!r}, line {lineno}: missing USE flag(s): {line!r}')
                continue
            d[a.key].append(misc.chunked_data(a, *split_negations(l[1:])))

        return ImmutableDict((k, misc._build_cp_atom_payload(v, atom(k)))
                             for k, v in d.items())

    def _parse_use(self, data):
        c = misc.ChunkedDataDict()
        data = (x[0] for x in data)
        neg, pos = split_negations(data)
        if neg or pos:
            c.add_bare_global(neg, pos)
        c.freeze()
        return c

    @load_property("use.force", allow_recurse=True)
    def use_force(self, data):
        return self._parse_use(data)

    @load_property("use.stable.force", allow_recurse=True,
                   eapi_optional='profile_stable_use')
    def use_stable_force(self, data):
        return self._parse_use(data)

    @load_property("package.use.force", allow_recurse=True)
    def pkg_use_force(self, data):
        return self._parse_package_use(data)

    @load_property("package.use.stable.force", allow_recurse=True,
                   eapi_optional='profile_stable_use')
    def pkg_use_stable_force(self, data):
        return self._parse_package_use(data)

    @load_property("use.mask", allow_recurse=True)
    def use_mask(self, data):
        return self._parse_use(data)

    @load_property("use.stable.mask", allow_recurse=True,
                   eapi_optional='profile_stable_use')
    def use_stable_mask(self, data):
        return self._parse_use(data)

    @load_property("package.use.mask", allow_recurse=True)
    def pkg_use_mask(self, data):
        return self._parse_package_use(data)

    @load_property("package.use.stable.mask", allow_recurse=True,
                   eapi_optional='profile_stable_use')
    def pkg_use_stable_mask(self, data):
        return self._parse_package_use(data)

    @klass.jit_attr
    def masked_use(self):
        c = self.use_mask
        if self.pkg_use_mask:
            c = c.clone(unfreeze=True)
            c.update_from_stream(
                chain.from_iterable(self.pkg_use_mask.values()))
            c.freeze()
        return c

    @klass.jit_attr
    def stable_masked_use(self):
        c = self.use_mask.clone(unfreeze=True)
        if self.use_stable_mask:
            c.merge(self.use_stable_mask)
        if self.pkg_use_mask:
            c.update_from_stream(
                chain.from_iterable(self.pkg_use_mask.values()))
        if self.pkg_use_stable_mask:
            c.update_from_stream(
                chain.from_iterable(self.pkg_use_stable_mask.values()))
        c.freeze()
        return c

    @klass.jit_attr
    def forced_use(self):
        c = self.use_force
        if self.pkg_use_force:
            c = c.clone(unfreeze=True)
            c.update_from_stream(
                chain.from_iterable(self.pkg_use_force.values()))
            c.freeze()
        return c

    @klass.jit_attr
    def stable_forced_use(self):
        c = self.use_force.clone(unfreeze=True)
        if self.use_stable_force:
            c.merge(self.use_stable_force)
        if self.pkg_use_force:
            c.update_from_stream(
                chain.from_iterable(self.pkg_use_force.values()))
        if self.pkg_use_stable_force:
            c.update_from_stream(
                chain.from_iterable(self.pkg_use_stable_force.values()))
        c.freeze()
        return c

    @load_property('make.defaults', read_func=None, fallback=None)
    def make_defaults(self, data):
        d = {}
        if data is not None:
            d.update(read_bash_dict(data[0]))
        return ImmutableDict(d)

    @load_property('make.defaults', read_func=None, fallback=None)
    def default_env(self, data):
        rendered = _make_incrementals_dict()
        for parent in self.parents:
            rendered.update(parent.default_env.items())

        if data is not None:
            data = read_bash_dict(data[0], vars_dict=rendered)
            rendered.update(data.items())
        return ImmutableDict(rendered)

    @klass.jit_attr
    def bashrc(self):
        path = pjoin(self.path, "profile.bashrc")
        if os.path.exists(path):
            return local_source(path)
        return None

    @load_property('eapi', fallback='0')
    def eapi(self, data):
        # handle fallback
        if isinstance(data, str):
            return get_eapi(data)

        try:
            line, lineno, relpath = next(data)
        except StopIteration:
            relpath = pjoin(self.name, 'eapi')
            logger.error(f'{relpath!r}: empty file')
            return get_eapi('0')

        try:
            next(data)
            logger.error(f'{relpath!r}: multiple lines detected')
        except StopIteration:
            pass

        eapi_str = line.strip()
        if eapi_str not in EAPI.known_eapis:
            logger.error(f'{relpath!r}: unknown EAPI {eapi_str!r}')
        return get_eapi(eapi_str)

    eapi_atom = klass.alias_attr("eapi.atom_kls")

    @klass.jit_attr
    def repoconfig(self):
        return self._load_repoconfig_from_path(self.path)

    @staticmethod
    def _load_repoconfig_from_path(path):
        path = abspath(path)
        # strip '/' so we don't get '/usr/portage' == ('', 'usr', 'portage')
        chunks = path.lstrip('/').split('/')
        try:
            pindex = max(idx for idx, x in enumerate(chunks) if x == 'profiles')
        except ValueError:
            # not in a repo...
            return None
        repo_path = pjoin('/', *chunks[:pindex])
        return repo_objs.RepoConfig(repo_path)

    @classmethod
    def _autodetect_and_create(cls, path):
        repo_config = cls._load_repoconfig_from_path(path)

        # note while this else seems pointless, we do it this
        # way so that we're not passing an arg unless needed- instance
        # caching is a bit overprotective, even if pms_strict defaults to True,
        # cls(path) is not cls(path, pms_strict=True)

        if repo_config is not None and 'pms' not in repo_config.profile_formats:
            profile = cls(path, pms_strict=False)
        else:
            profile = cls(path)

        # optimization to avoid re-parsing what we already did.
        object.__setattr__(profile, '_repoconfig', repo_config)
        return profile


class EmptyRootNode(ProfileNode):

    __inst_caching__ = True

    parents = ()
    deprecated = None
    pkg_use = masked_use = stable_masked_use = forced_use = stable_forced_use = misc.ChunkedDataDict()
    forced_use.freeze()
    pkg_use_force = pkg_use_mask = ImmutableDict()
    pkg_provided = system = profile_set = ((), ())


class ProfileStack:

    _node_kls = ProfileNode

    def __init__(self, profile):
        self.profile = profile
        self.node = self._node_kls._autodetect_and_create(profile)

    @property
    def arch(self):
        return self.default_env.get("ARCH")

    deprecated = klass.alias_attr("node.deprecated")
    eapi = klass.alias_attr("node.eapi")
    name = klass.alias_attr("node.name")

    @klass.jit_attr
    def stack(self):
        def f(node):
            for path, line, lineno in node.parent_paths:
                try:
                    x = self._node_kls._autodetect_and_create(path)
                except ProfileError as e:
                    repo_id = node.repoconfig.repo_id
                    logger.error(
                        f"repo {repo_id!r}: '{self.name}/parent' (line {lineno}), "
                        f'bad profile parent {line!r}: {e.error}'
                    )
                    continue
                for y in f(x):
                    yield y
            yield node
        return tuple(f(self.node))

    @klass.jit_attr
    def _system_profile(self):
        """User-selected system profile.

        This should map directly to the profile linked to /etc/portage/make.profile.
        """
        # prefer main system profile; otherwise, fallback to custom user profile
        for profile in reversed(self.stack):
            if not isinstance(profile, UserProfileNode):
                break
        return profile

    def _collapse_use_dict(self, attr):
        stack = (getattr(x, attr) for x in self.stack)
        d = misc.ChunkedDataDict()
        for mapping in stack:
            d.merge(mapping)
        d.freeze()
        return d

    @klass.jit_attr
    def forced_use(self):
        return self._collapse_use_dict("forced_use")

    @klass.jit_attr
    def masked_use(self):
        return self._collapse_use_dict("masked_use")

    @klass.jit_attr
    def stable_forced_use(self):
        return self._collapse_use_dict("stable_forced_use")

    @klass.jit_attr
    def stable_masked_use(self):
        return self._collapse_use_dict("stable_masked_use")

    @klass.jit_attr
    def pkg_use(self):
        return self._collapse_use_dict("pkg_use")

    def _collapse_generic(self, attr, clear=False):
        s = set()
        for node in self.stack:
            val = getattr(node, attr)
            if clear and len(val) > 2 and val[2]:
                s.clear()
            s.difference_update(val[0])
            s.update(val[1])
        return s

    @klass.jit_attr
    def default_env(self):
        d = dict(self.node.default_env.items())
        for incremental in const.incrementals:
            v = d.pop(incremental, '').split()
            if v:
                if incremental in const.incrementals_unfinalized:
                    d[incremental] = tuple(v)
                else:
                    v = misc.incremental_expansion(
                        v, msg_prefix=f"While expanding {incremental}, value {v!r}: ")
                    if v:
                        d[incremental] = tuple(v)
        return ImmutableDict(d.items())

    @property
    def profile_only_variables(self):
        if "PROFILE_ONLY_VARIABLES" in const.incrementals:
            return frozenset(self.default_env.get("PROFILE_ONLY_VARIABLES", ()))
        return frozenset(self.default_env.get("PROFILE_ONLY_VARIABLES", "").split())

    @klass.jit_attr
    def use_expand(self):
        """USE_EXPAND variables defined by the profile."""
        if "USE_EXPAND" in const.incrementals:
            return frozenset(self.default_env.get("USE_EXPAND", ()))
        return frozenset(self.default_env.get("USE_EXPAND", "").split())

    @klass.jit_attr
    def use(self):
        """USE flag settings for the profile."""
        return tuple(list(self.default_env.get('USE', ())) + list(self.expand_use()))

    def expand_use(self, env=None):
        """Expand USE_EXPAND settings to USE flags."""
        if env is None:
            env = self.default_env

        use = []
        for u in self.use_expand:
            value = env.get(u)
            if value is None:
                continue
            u2 = u.lower() + '_'
            use.extend(u2 + x for x in value.split())
        return tuple(use)

    @property
    def use_expand_hidden(self):
        if "USE_EXPAND_HIDDEN" in const.incrementals:
            return frozenset(self.default_env.get("USE_EXPAND_HIDDEN", ()))
        return frozenset(self.default_env.get("USE_EXPAND_HIDDEN", "").split())

    @property
    def iuse_implicit(self):
        if "IUSE_IMPLICIT" in const.incrementals:
            return frozenset(self.default_env.get("IUSE_IMPLICIT", ()))
        return frozenset(self.default_env.get("IUSE_IMPLICIT", "").split())

    @property
    def use_expand_implicit(self):
        if "USE_EXPAND_IMPLICIT" in const.incrementals:
            return frozenset(self.default_env.get("USE_EXPAND_IMPLICIT", ()))
        return frozenset(self.default_env.get("USE_EXPAND_IMPLICIT", "").split())

    @property
    def use_expand_unprefixed(self):
        if "USE_EXPAND_UNPREFIXED" in const.incrementals:
            return frozenset(self.default_env.get("USE_EXPAND_UNPREFIXED", ()))
        return frozenset(self.default_env.get("USE_EXPAND_UNPREFIXED", "").split())

    @klass.jit_attr
    def iuse_effective(self):
        iuse_effective = []

        # EAPI 5 and above allow profile defined IUSE injection (see PMS)
        if self._system_profile.eapi.options.profile_iuse_injection:
            iuse_effective.extend(self.iuse_implicit)
            for v in self.use_expand_implicit.intersection(self.use_expand_unprefixed):
                iuse_effective.extend(self.default_env.get("USE_EXPAND_VALUES_" + v, "").split())
            for v in self.use_expand.intersection(self.use_expand_implicit):
                for x in self.default_env.get("USE_EXPAND_VALUES_" + v, "").split():
                    iuse_effective.append(v.lower() + "_" + x)
        else:
            iuse_effective.extend(self._system_profile.repoconfig.known_arches)
            for v in self.use_expand:
                for x in self.default_env.get("USE_EXPAND_VALUES_" + v, "").split():
                    iuse_effective.append(v.lower() + "_" + x)

        return frozenset(iuse_effective)

    @klass.jit_attr
    def provides_repo(self):
        # delay importing to avoid circular imports
        from .repository import ProvidesRepo
        pkgs = self._collapse_generic('pkg_provided')
        try:
            arches = self._system_profile.repoconfig.known_arches
        except AttributeError:
            # TODO: repoconfig is None when using fake repos
            arches = ()
        return ProvidesRepo(pkgs, arches)

    @klass.jit_attr
    def masks(self):
        return frozenset(chain(self._collapse_generic("masks")))

    @klass.jit_attr
    def unmasks(self):
        return frozenset(self._collapse_generic('unmasks'))

    @klass.jit_attr
    def pkg_deprecated(self):
        return frozenset(chain(self._collapse_generic("pkg_deprecated")))

    @klass.jit_attr
    def keywords(self):
        return tuple(chain.from_iterable(x.keywords for x in self.stack))

    @klass.jit_attr
    def accept_keywords(self):
        return tuple(chain.from_iterable(x.accept_keywords for x in self.stack))

    def _incremental_masks(self, stack_override=None):
        if stack_override is None:
            stack_override = self.stack
        return tuple(node.masks for node in stack_override if any(node.masks))

    def _incremental_unmasks(self, stack_override=None):
        if stack_override is None:
            stack_override = self.stack
        return tuple(node.unmasks for node in stack_override if any(node.unmasks))

    @klass.jit_attr
    def bashrcs(self):
        return tuple(x.bashrc for x in self.stack if x.bashrc is not None)

    bashrc = klass.alias_attr("bashrcs")
    path = klass.alias_attr("node.path")

    @klass.jit_attr
    def system(self):
        return frozenset(self._collapse_generic('system', clear=True))

    @klass.jit_attr
    def profile_set(self):
        return frozenset(self._collapse_generic('profile_set', clear=True))


class OnDiskProfile(ProfileStack):

    pkgcore_config_type = ConfigHint(
        {'basepath': 'str', 'profile': 'str'},
        required=('basepath', 'profile'),
        typename='profile',
    )

    def __init__(self, basepath, profile, load_profile_base=True):
        super().__init__(pjoin(basepath, profile))
        self.basepath = basepath
        self.load_profile_base = load_profile_base

    @staticmethod
    def split_abspath(path):
        path = abspath(path)
        # filter's heavy, but it handles '/' while also
        # suppressing the leading '/'
        chunks = [x for x in path.split("/") if x]
        try:
            # poor mans rindex.
            pbase = max(idx for idx, x in enumerate(chunks) if x == 'profiles')
        except ValueError:
            # no base found.
            return None
        return pjoin("/", *chunks[:pbase+1]), '/'.join(chunks[pbase+1:])

    @classmethod
    def from_abspath(cls, path):
        vals = cls.split_abspath(path)
        if vals is not None:
            vals = cls(load_profile_base=True, *vals)
        return vals

    @klass.jit_attr
    def stack(self):
        l = ProfileStack.stack.function(self)
        if self.load_profile_base:
            l = (EmptyRootNode._autodetect_and_create(self.basepath),) + l
        return l

    @klass.jit_attr
    def _incremental_masks(self):
        stack = self.stack
        if self.load_profile_base:
            stack = stack[1:]
        return ProfileStack._incremental_masks(self, stack_override=stack)

    @klass.jit_attr
    def _incremental_unmasks(self):
        stack = self.stack
        if self.load_profile_base:
            stack = stack[1:]
        return ProfileStack._incremental_unmasks(self, stack_override=stack)


class UserProfileNode(ProfileNode):

    parent_node_kls = ProfileNode

    def __init__(self, path, parent_path):
        self.override_path = pjoin(path, parent_path)
        super().__init__(path, pms_strict=False)

    @klass.jit_attr
    def parents(self):
        return (ProfileNode(self.override_path),)

    @klass.jit_attr
    def parent_paths(self):
        return ((self.override_path, None, None),)


class UserProfile(OnDiskProfile):

    pkgcore_config_type = ConfigHint(
        {'user_path': 'str', 'parent_path': 'str', 'parent_profile': 'str'},
        required=('user_path', 'parent_path', 'parent_profile'),
        typename='profile',
    )

    def __init__(self, user_path, parent_path, parent_profile, load_profile_base=True):
        super().__init__(parent_path, parent_profile, load_profile_base)
        self.node = UserProfileNode(user_path, pjoin(parent_path, parent_profile))



"""
pkgset based around loading a list of atoms from a world file
"""

__all__ = ("FileList", "WorldFile")

from snakeoil import klass
from snakeoil.fileutils import AtomicWriteFile, readlines_ascii

from .. import os_data
from ..config import errors
from ..config.hint import ConfigHint
from ..ebuild import const
from ..ebuild.atom import atom
from ..log import logger
from ..package.errors import InvalidDependency


class FileList:
    pkgcore_config_type = ConfigHint({'location': 'str'}, typename='pkgset')
    error_on_subsets = True

    def __init__(self, location, gid=os_data.portage_gid, mode=0o644):
        self.path = location
        self.gid = gid
        self.mode = mode
        # note that _atoms is generated on the fly.

    @klass.jit_attr
    def _atoms(self):
        try:
            s = set()
            for x in readlines_ascii(self.path, True):
                if not x or x.startswith("#"):
                    continue
                elif x.startswith("@"):
                    if self.error_on_subsets:
                        raise ValueError(
                            "set %s isn't a valid atom in pkgset %r" %
                            (x, self.path))
                    logger.warning(
                        "set item %r found in pkgset %r: it will be "
                        "wiped on update since portage/pkgcore store set items "
                        "in a separate way", x[1:], self.path)
                    continue
                s.add(atom(x))
        except InvalidDependency as e:
            raise errors.ParsingError("parsing %r" % self.path, exception=e) from e

        return s

    def __iter__(self):
        return iter(self._atoms)

    def __len__(self):
        return len(self._atoms)

    def __contains__(self, key):
        return key in self._atoms

    def add(self, atom_inst):
        self._atoms.add(atom_inst)

    def remove(self, atom_inst):
        self._atoms.remove(atom_inst)

    def flush(self):
        f = None
        try:
            f = AtomicWriteFile(self.path, gid=self.gid, perms=self.mode)
            f.write("\n".join(str(x) for x in sorted(self._atoms)))
            f.close()
        except:
            if f is not None:
                f.discard()
            raise


class WorldFile(FileList):
    """Set of packages contained in the world file."""
    pkgcore_config_type = ConfigHint(typename='pkgset')
    error_on_subsets = False

    def __init__(self, location=const.WORLD_FILE,
                 gid=os_data.portage_gid, mode=0o644):
        FileList.__init__(self, location, gid=gid, mode=mode)

    def add(self, atom_inst):
        self._modify(atom_inst, FileList.add)

    def remove(self, atom_inst):
        self._modify(atom_inst, FileList.remove)

    def _modify(self, atom_inst, func):
        if atom_inst.slot:
            for slot in atom_inst.slot:
                if slot == '0':
                    new_atom_inst = atom(atom_inst.key)
                else:
                    new_atom_inst = atom(atom_inst.key + ":" + slot)
                func(self, new_atom_inst)
        else:
            atom_inst = atom(atom_inst.key)
            func(self, atom_inst)



"""package with its metadata accessible (think 'no longer abstract')"""

__all__ = ("DeriveMetadataKls", "factory", "package")

from weakref import WeakValueDictionary

from snakeoil import klass

from ..ebuild import cpv
from ..ebuild.atom import atom
from . import base


def DeriveMetadataKls(original_kls):
    if getattr(original_kls, "_derived_metadata_kls", False):
        return original_kls

    class package(original_kls, metaclass=base.DynamicGetattrSetter):
        _derived_metadata_kls = True
        built = False
        __slots__ = ("_parent", "data", "_domain")
        try:
            __doc__ = "package class with metadata bound to it for attribute " \
                "generation\n\n" + \
                     "\n".join(x.lstrip()
                          for x in original_kls.__doc__.split("\n")
                          if ":ivar" in x or ":cvar" in x)
            __doc__ += "\n:ivar repo: parent repository"
        except AttributeError:
            # wee, must be in -OO mode.
            __doc__ = None

        immutable = True
        package_is_real = True

        def __init__(self, parent_repository, *args, **kwds):
            f"""wrapper for {original_kls}.__init__
            
            See {original_kls}.__init__ for allowed args/kwds, they're passed
            directly to it.

            :param parent_repository: parent repository this package belongs to
            :type parent_repository: :obj:`pkgcore.repository.prototype.tree`
                instance
            """
            super().__init__(*args, **kwds)
            object.__setattr__(self, '_parent',  parent_repository)

        @base.DynamicGetattrSetter.register
        def data(self):
            """internal hook func to get the packages metadata"""
            return self._fetch_metadata()

        repo = klass.alias_attr("_parent._parent_repo")

        def release_cached_data(self, all=False):
            for x in self._get_attr:
                try:
                    object.__delattr__(self, x)
                except AttributeError:
                    pass

            if all:
                try:
                    object.__delattr__(self, 'data')
                except AttributeError:
                    pass

        @property
        def slotted_atom(self):
            return atom(f'{self.key}:{self.slot}')

        def _fetch_metadata(self):
            """Pull the metadata for this package.

            Must be overridden in derivatives.
            """
            raise NotImplementedError

        def add_format_triggers(self, op_inst, format_op_inst, engine_inst):
            pass

    return package

package = DeriveMetadataKls(cpv.VersionedCPV)

class factory:
    """package generator

    does weakref caching per repository

    :cvar child_class: callable to generate packages
    """

    child_class = package

    def __init__(self, parent_repo):
        self._parent_repo = parent_repo
        self._cached_instances = WeakValueDictionary()

    def new_package(self, *args):
        """generate a new package instance"""
        inst = self._cached_instances.get(args)
        if inst is None:
            inst = self._cached_instances[args] = self.child_class(self, *args)
        return inst

    def __call__(self, *args, **kwds):
        return self.new_package(*args, **kwds)

    def clear(self):
        """wipe the weakref cache of packages instances"""
        self._cached_instances.clear()

    def _get_metadata(self, *args):
        """Pulls metadata from the repo/cache/wherever.

        Must be overridden in derivatives.
        """
        raise NotImplementedError

    def _update_metadata(self, *args):
        """Updates metadata in the repo/cache/wherever.

        Must be overridden in derivatives.
        """
        raise NotImplementedError

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_cached_instances']
        return d

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        self.__dict__['_cached_instances'] = WeakValueDictionary()



from pkgcore.package import base, metadata
from snakeoil.test import TestCase


def make_pkg_kls(attrs=(), callbacks={}):

    class simple_pkg(base.base):
        _get_attr = callbacks
        __slots__ = ("_args", "_kwds", "_data", "_fetch_called",) + tuple(attrs)

        def __init__(self, *args, **kwds):
            self._args = args
            self._data = kwds.pop("data", {})
            self._kwds = kwds
            self._fetch_called = False

        __setattr__ = object.__setattr__

    class metadata_pkg(metadata.DeriveMetadataKls(simple_pkg)):

        __slots__ = ()
        def _fetch_metadata(self):
            self._fetch_called = True
            return self._data

    return metadata_pkg


class TestMetadataPackage(TestCase):

    def test_init(self):
        class repo:
            _parent_repo = "foon"

        kls = make_pkg_kls()
        o = kls(repo, "monkeys", dar=1)
        self.assertEqual(o._parent, repo)
        self.assertEqual(o.repo, "foon")
        self.assertEqual(o._args, ("monkeys",))
        self.assertEqual(o._kwds, {"dar":1})
        self.assertEqual(o._fetch_called, False)

    def test_getdata(self):
        kls = make_pkg_kls()
        o = kls(None, data={'a': 'b'})
        self.assertEqual(o.data, {'a': 'b'})



"""
filesystem entry abstractions
"""

import fnmatch
import stat
from os.path import abspath, basename, dirname, realpath
from os.path import sep as path_seperator

from snakeoil import klass
from snakeoil.chksum import get_chksums, get_handlers
from snakeoil.compatibility import cmp
from snakeoil.currying import post_curry, pretty_docs
from snakeoil.data_source import local_source
from snakeoil.mappings import LazyFullValLoadDict
from snakeoil.osutils import normpath, pjoin

# goofy set of classes representating the fs objects pkgcore knows of.

__all__ = [
    "fsFile", "fsDir", "fsSymlink", "fsDev", "fsFifo"]
__all__.extend(
    f"is{x}" for x in ("dir", "reg", "sym", "fifo", "dev", "fs_obj"))

# following are used to generate appropriate __init__, wiped from the
# namespace at the end of the module

_fs_doc = {
    "mode":""":keyword mode: int, the mode of this entry.  """
        """required if strict is set""",
    "mtime":""":keyword mtime: long, the mtime of this entry.  """
        """required if strict is set""",
    "uid":""":keyword uid: int, the uid of this entry.  """
        """required if strict is set""",
    "gid":""":keyword gid: int, the gid of this entry.  """
        """required if strict is set""",
}

def gen_doc_additions(init, slots):
    if init.__doc__ is None:
        d = \
"""
:param location: location (real or intended) for this entry
:param strict: is this fully representative of the entry, or only partially
:raise KeyError: if strict is enabled, and not all args are passed in
""".split("\n")
    else:
        d = init.__doc__.split("\n")
    init.__doc__ = "\n".join(k.lstrip() for k in d) + \
        "\n".join(_fs_doc[k] for k in _fs_doc if k in slots)


class fsBase:

    """base class, all extensions must derive from this class"""
    __slots__ = ("location", "mtime", "mode", "uid", "gid")
    __attrs__ = __slots__
    __default_attrs__ = {}

    locals().update((x.replace("is", "is_"), False) for x in
        __all__ if x.startswith("is") and x.islower() and not
            x.endswith("fs_obj"))

    klass.inject_richcmp_methods_from_cmp(locals())
    klass.inject_immutable_instance(locals())

    def __init__(self, location, strict=True, **d):

        d["location"] = normpath(location)

        s = object.__setattr__
        if strict:
            for k in self.__attrs__:
                s(self, k, d[k])
        else:
            for k, v in d.items():
                s(self, k, v)
    gen_doc_additions(__init__, __attrs__)

    def change_attributes(self, **kwds):
        d = {x: getattr(self, x)
             for x in self.__attrs__ if hasattr(self, x)}
        d.update(kwds)
        # split location out
        location = d.pop("location")
        if not location.startswith(path_seperator):
            location = abspath(location)
        d["strict"] = False
        return self.__class__(location, **d)

    def __getattr__(self, attr):
        # we would only get called if it doesn't exist.
        if attr not in self.__attrs__:
            raise AttributeError(self, attr)
        obj = self.__default_attrs__.get(attr)
        if not callable(obj):
            return obj
        return obj(self)

    def __hash__(self):
        return hash(self.location)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.location == other.location

    def __ne__(self, other):
        return not self == other

    def realpath(self, cache=None):
        """calculate the abspath/canonicalized path for this entry, returning
        a new instance if the path differs.

        :keyword cache: Either None (no cache), or a data object of path->
          resolved.  Currently unused, but left in for forwards compatibility
        """
        new_path = realpath(self.location)
        if new_path == self.location:
            return self
        return self.change_attributes(location=new_path)

    @property
    def basename(self):
        return basename(self.location)

    @property
    def dirname(self):
        return dirname(self.location)

    def fnmatch(self, pattern):
        return fnmatch.fnmatch(self.location, pattern)

    def __cmp__(self, other):
        return cmp(self.location, other.location)

    def __str__(self):
        return self.location


class _LazyChksums(LazyFullValLoadDict):
    __slots__ = ()


class fsFile(fsBase):

    """file class"""

    __slots__ = ("chksums", "data", "dev", "inode")
    __attrs__ = fsBase.__attrs__ + __slots__
    __default_attrs__ = {"mtime":0, 'dev':None, 'inode':None}

    is_reg = True

    def __init__(self, location, chksums=None, data=None, **kwds):
        """
        :param chksums: dict of checksums, key chksum_type: val hash val.
            See :obj:`snakeoil.chksum`.
        """
        assert 'data_source' not in kwds
        if data is None:
            data = local_source(location)
        kwds["data"] = data

        if chksums is None:
            # this can be problematic offhand if the file is modified
            # but chksum not triggered
            chf_types = kwds.pop("chf_types", None)
            if chf_types is None:
                chf_types = tuple(get_handlers())
            chksums = _LazyChksums(chf_types, self._chksum_callback)
        kwds["chksums"] = chksums
        fsBase.__init__(self, location, **kwds)
    gen_doc_additions(__init__, __slots__)

    def __repr__(self):
        return f"file:{self.location}"

    data_source = klass.alias_attr("data")

    def _chksum_callback(self, chfs):
        return list(zip(chfs, get_chksums(self.data, *chfs)))

    def change_attributes(self, **kwds):
        if 'data' in kwds and ('chksums' not in kwds and
            isinstance(self.chksums, _LazyChksums)):
            kwds['chksums'] = None
        return fsBase.change_attributes(self, **kwds)

    def _can_be_hardlinked(self, other):
        if not other.is_reg:
            return False

        if None in (self.inode, self.dev):
            return False

        for attr in ('dev', 'inode', 'uid', 'gid', 'mode', 'mtime'):
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True


class fsDir(fsBase):

    """dir class"""

    __slots__ = ()
    is_dir = True

    def __repr__(self):
        return f"dir:{self.location}"


class fsLink(fsBase):

    """symlink class"""

    __slots__ = ("target",)
    __attrs__ = fsBase.__attrs__ + __slots__
    is_sym = True

    def __init__(self, location, target, **kwargs):
        """
        :param target: string, filepath of the symlinks target
        """
        kwargs["target"] = target
        fsBase.__init__(self, location, **kwargs)
    gen_doc_additions(__init__, __slots__)

    def change_attributes(self, **kwds):
        d = {x: getattr(self, x)
             for x in self.__attrs__ if hasattr(self, x)}
        d.update(kwds)
        # split location out
        location = d.pop("location")
        if not location.startswith(path_seperator):
            location = abspath(location)
        target = d.pop("target")
        d["strict"] = False
        return self.__class__(location, target, **d)

    @property
    def resolved_target(self):
        if self.target.startswith("/"):
            return self.target
        return normpath(pjoin(self.location, '../', self.target))

    def __cmp__(self, other):
        c = cmp(self.location, other.location)
        if c:
            return c
        if isinstance(other, self.__class__):
            return cmp(self.target, other.target)
        return 0

    def __str__(self):
        return f'{self.location} -> {self.target}'

    def __repr__(self):
        return f"symlink:{self.location}->{self.target}"


fsSymlink = fsLink


class fsDev(fsBase):

    """dev class (char/block objects)"""

    __slots__ = ("major", "minor")
    __attrs__ = fsBase.__attrs__ + __slots__
    __default_attrs__ = {"major":-1, "minor":-1}
    is_dev = True

    def __init__(self, path, major=-1, minor=-1, **kwds):
        if kwds.get("strict", True):
            if major == -1 or minor == -1:
                raise TypeError(
                   "major/minor must be specified and positive ints")
            if not stat.S_IFMT(kwds["mode"]):
                raise TypeError(
                    "mode %o: must specify the device type (got %o)" % (
                        kwds["mode"], stat.S_IFMT(kwds["mode"])))
            kwds["major"] = major
            kwds["minor"] = minor
        else:
            if major != -1:
                major = int(major)
                if major < 0:
                    raise TypeError(
                       "major/minor must be specified and positive ints")
                kwds["major"] = major

            if minor != -1:
                minor = int(minor)
                if minor < 0:
                    raise TypeError(
                       "major/minor must be specified and positive ints")
                kwds["minor"] = minor

        fsBase.__init__(self, path, **kwds)

    def __repr__(self):
        return f"device:{self.location}"


def get_major_minor(stat_inst):
    """get major/minor from a stat instance
    :return: major,minor tuple of ints
    """
    return ( stat_inst.st_rdev >> 8 ) & 0xff, stat_inst.st_rdev & 0xff


class fsFifo(fsBase):

    """fifo class (socket objects)"""

    __slots__ = ()
    is_fifo = True

    def __repr__(self):
        return f"fifo:{self.location}"

def mk_check(name):
    return pretty_docs(post_curry(getattr, 'is_' + name, False),
        extradocs=("return True if obj is an instance of :obj:`%s`, else False" % name),
        name=("is" +name)
        )

isdir    = mk_check('dir')
isreg    = mk_check('reg')
issym    = mk_check('sym')
isfifo   = mk_check('fifo')
isdev    = mk_check('dev')
isfs_obj = pretty_docs(post_curry(isinstance, fsBase), name='isfs_obj',
    extradocs='return True if obj is an fsBase derived object')

del gen_doc_additions, mk_check



from snakeoil.compatibility import IGNORED_EXCEPTIONS

from ..package.errors import MetadataException
from ..util.thread_pool import map_async


def regen_iter(iterable, regen_func, observer):
    for pkg in iterable:
        try:
            regen_func(pkg)
        except IGNORED_EXCEPTIONS as e:
            if isinstance(e, KeyboardInterrupt):
                return
            raise
        except MetadataException:
            # handled at a higher level by scanning for metadata masked pkgs
            # after regen has completed
            pass
        except Exception as e:
            yield pkg, e


def regen_repository(repo, pkgs, observer, threads=1, pkg_attr='keywords', **kwargs):
    helpers = []

    def _get_repo_helper():
        if not hasattr(repo, '_regen_operation_helper'):
            return lambda pkg: getattr(pkg, 'keywords')
        # for an actual helper, track it and invoke .finish if it exists.
        helper = repo._regen_operation_helper(**kwargs)
        helpers.append(helper)
        return helper

    def get_args():
        return (_get_repo_helper(), observer)

    errors = map_async(pkgs, regen_iter, threads=threads, per_thread_args=get_args)

    # yield any errors that occurred during metadata generation
    yield from errors



import os
import textwrap
from unittest import mock

from pkgcore.ebuild import eclass_cache
from pkgcore.ebuild import errors as ebuild_errors
from pkgcore.ebuild import repository, restricts
from pkgcore.ebuild.atom import atom
from pkgcore.repository import errors
from snakeoil.fileutils import touch
from snakeoil.osutils import ensure_dirs, pjoin
from snakeoil.test.mixins import TempDirMixin


class TestUnconfiguredTree(TempDirMixin):

    def mk_tree(self, path, *args, **kwds):
        eclasses = kwds.pop('eclass_cache', None)
        if eclasses is None:
            epath = pjoin(path, 'eclass')
            ensure_dirs(epath)
            eclasses = eclass_cache.cache(epath)
        ensure_dirs(pjoin(path, 'profiles'))
        return repository.UnconfiguredTree(path, eclass_cache=eclasses, *args, **kwds)

    def setUp(self):
        TempDirMixin.setUp(self)
        self.pdir = pjoin(self.dir, 'profiles')
        ensure_dirs(self.pdir)

        # silence missing masters warnings
        ensure_dirs(pjoin(self.dir, 'metadata'))
        with open(pjoin(self.dir, 'metadata', 'layout.conf'), 'w') as f:
            f.write('masters =\n')

    def test_basics(self):
        repo = self.mk_tree(self.dir)
        # Just make sure these do not raise.
        self.assertTrue(str(repo))
        self.assertTrue(repr(repo))

        self.assertRaises(
            errors.InitializationError,
            self.mk_tree, pjoin(self.dir, 'missing'))

        with open(pjoin(self.dir, 'random'), 'w') as f:
            f.write('random')
        self.assertRaises(
            errors.InitializationError,
            self.mk_tree, pjoin(self.dir, 'random'))

    def test_thirdpartymirrors(self):
        with open(pjoin(self.pdir, 'thirdpartymirrors'), 'w') as f:
            f.write(textwrap.dedent('''\
                spork http://sporks/ http://moresporks/
                foon foon://foons/
            '''))
        mirrors = self.mk_tree(self.dir).mirrors
        self.assertEqual(['foon', 'spork'], sorted(mirrors))
        self.assertEqual(
            ['http://moresporks/', 'http://sporks/'],
            sorted(mirrors['spork']))
        with open(pjoin(self.pdir, 'thirdpartymirrors'), 'w') as f:
            f.write("foon  dar\n")
        self.assertEqual(list(self.mk_tree(self.dir).mirrors.keys()), ['foon'])

    def test_repo_id(self):
        dir1 = pjoin(self.dir, '1')
        os.mkdir(dir1, 0o755)
        repo = self.mk_tree(dir1)
        self.assertEqual(repo.repo_id, dir1)
        dir2 = pjoin(self.dir, '2')
        ensure_dirs(pjoin(dir2, 'profiles'))
        with open(pjoin(dir2, 'profiles', 'repo_name'), 'w') as f:
            f.write('testrepo\n')
        repo = self.mk_tree(dir2)
        self.assertEqual('testrepo', repo.repo_id)

    def test_licenses(self):
        licenses = ('GPL-2', 'GPL-3+', 'BSD')
        ensure_dirs(pjoin(self.dir, 'licenses'))
        for license in licenses:
            touch(pjoin(self.dir, 'licenses', license))
        repo = self.mk_tree(self.dir)
        self.assertEqual(sorted(repo.licenses), sorted(licenses))

    def test_masters(self):
        repo = self.mk_tree(self.dir)
        self.assertEqual(repo.masters, ())

    def test_path_restrict(self):
        repo_dir = pjoin(self.dir, 'repo')
        sym_repo_dir = pjoin(self.dir, 'sym_repo')
        os.symlink(repo_dir, sym_repo_dir)

        ensure_dirs(pjoin(repo_dir, 'profiles'))
        with open(pjoin(repo_dir, 'profiles', 'repo_name'), 'w') as f:
            f.write('testrepo\n')
        ensure_dirs(pjoin(repo_dir, 'cat', 'foo'))
        ensure_dirs(pjoin(repo_dir, 'cat', 'bar'))
        ensure_dirs(pjoin(repo_dir, 'tac', 'oof'))
        touch(pjoin(repo_dir, 'skel.ebuild'))
        touch(pjoin(repo_dir, 'cat', 'foo', 'Manifest'))
        ebuilds = (
            pjoin(repo_dir, 'cat', 'foo', 'foo-1.ebuild'),
            pjoin(repo_dir, 'cat', 'foo', 'foo-2.ebuild'),
            pjoin(repo_dir, 'cat', 'bar', 'bar-1.ebuild'),
            pjoin(repo_dir, 'tac', 'oof', 'oof-1.ebuild'),
        )
        for ebuild in ebuilds:
            with open(ebuild, 'w') as f:
                f.write('SLOT=0\n')

        # specify repo category dirs
        with open(pjoin(repo_dir, 'profiles', 'categories'), 'w') as f:
            f.write('cat\n')
            f.write('tac\n')

        for d in (repo_dir, sym_repo_dir):
            repo = self.mk_tree(d)
            for path in (
                    self.dir,  # path not in repo
                    pjoin(repo.location, 'a'),  # nonexistent category dir
                    pjoin(repo.location, 'profiles'),  # non-category dir
                    pjoin(repo.location, 'skel.ebuild'),  # not in the correct cat/PN dir layout
                    pjoin(repo.location, 'cat', 'a'),  # nonexistent package dir
                    pjoin(repo.location, 'cat', 'foo', 'foo-0.ebuild'),  # nonexistent ebuild file
                    pjoin(repo.location, 'cat', 'foo', 'Manifest'),  # non-ebuild file
                    ):
                self.assertRaises(ValueError, repo.path_restrict, path)

            # repo dir
            restriction = repo.path_restrict(repo.location)
            self.assertEqual(len(restriction), 1)
            self.assertInstance(restriction[0], restricts.RepositoryDep)
            # matches all 4 ebuilds in the repo
            self.assertEqual(len(repo.match(restriction)), 4)

            # category dir
            restriction = repo.path_restrict(pjoin(repo.location, 'cat'))
            self.assertEqual(len(restriction), 2)
            self.assertInstance(restriction[1], restricts.CategoryDep)
            # matches all 3 ebuilds in the category
            self.assertEqual(len(repo.match(restriction)), 3)

            # relative category dir
            with mock.patch('os.getcwd', return_value=repo.location):
                restriction = repo.path_restrict('cat')
                self.assertEqual(len(restriction), 2)
                self.assertInstance(restriction[1], restricts.CategoryDep)
                # matches all 3 ebuilds in the category
                self.assertEqual(len(repo.match(restriction)), 3)

            # package dir
            restriction = repo.path_restrict(pjoin(repo.location, 'cat', 'foo'))
            self.assertEqual(len(restriction), 3)
            self.assertInstance(restriction[2], restricts.PackageDep)
            # matches both ebuilds in the package dir
            self.assertEqual(len(repo.match(restriction)), 2)

            # relative package dir
            with mock.patch('os.getcwd', return_value=repo.location):
                restriction = repo.path_restrict('cat/foo')
                self.assertEqual(len(restriction), 3)
                self.assertInstance(restriction[2], restricts.PackageDep)
                # matches both ebuilds in the package dir
                self.assertEqual(len(repo.match(restriction)), 2)

            # ebuild file
            restriction = repo.path_restrict(pjoin(repo.location, 'cat', 'foo', 'foo-1.ebuild'))
            self.assertEqual(len(restriction), 4)
            self.assertInstance(restriction[3], restricts.VersionMatch)
            # specific ebuild version match
            self.assertEqual(len(repo.match(restriction)), 1)

            # relative ebuild file path
            with mock.patch('os.getcwd', return_value=os.path.realpath(pjoin(repo.location, 'cat', 'foo'))):
                restriction = repo.path_restrict('./foo-1.ebuild')
                self.assertEqual(len(restriction), 4)
                self.assertInstance(restriction[3], restricts.VersionMatch)
                # specific ebuild version match
                self.assertEqual(len(repo.match(restriction)), 1)

    def test_categories_packages(self):
        ensure_dirs(pjoin(self.dir, 'cat', 'pkg'))
        ensure_dirs(pjoin(self.dir, 'empty', 'empty'))
        touch(pjoin(self.dir, 'cat', 'pkg', 'pkg-3.ebuild'))
        repo = self.mk_tree(self.dir)
        assert {'cat': (), 'empty': ()} == dict(repo.categories)
        assert {'cat': ('pkg',), 'empty': ('empty',)} == dict(repo.packages)
        assert {('cat', 'pkg'): ('3',), ('empty', 'empty'): ()} == dict(repo.versions)

    def test_package_mask(self):
        with open(pjoin(self.pdir, 'package.mask'), 'w') as f:
            f.write(textwrap.dedent('''\
                # lalala
                it-is/broken
                <just/newer-than-42
            '''))
        repo = self.mk_tree(self.dir)
        self.assertEqual(sorted([atom('it-is/broken'),
            atom('<just/newer-than-42')]),
            sorted(repo.pkg_masks))


class TestSlavedTree(TestUnconfiguredTree):

    def mk_tree(self, path, *args, **kwds):
        if path != self.dir:
            self.dir_slave = path
            self.dir_master = pjoin(os.path.dirname(path), os.path.basename(path) + 'master')
            ensure_dirs(self.dir_slave)
            ensure_dirs(self.dir_master)
            ensure_dirs(pjoin(self.dir_slave, 'profiles'))
            ensure_dirs(pjoin(self.dir_master, 'profiles'))

        eclasses = kwds.pop('eclass_cache', None)
        if eclasses is None:
            epath = pjoin(self.dir_master, 'eclass')
            ensure_dirs(epath)
            eclasses = eclass_cache.cache(epath)

        self.master_repo = repository.UnconfiguredTree(self.dir_master, eclass_cache=eclasses, *args, **kwds)
        masters = (self.master_repo,)
        return repository.UnconfiguredTree(self.dir_slave, eclass_cache=eclasses, masters=masters, *args, **kwds)

    def setUp(self):
        TempDirMixin.setUp(self)
        self.dir_orig = self.dir

        self.dir_master = pjoin(self.dir, 'master')
        self.dir_slave = pjoin(self.dir, 'slave')
        ensure_dirs(self.dir_master)
        ensure_dirs(self.dir_slave)

        ensure_dirs(pjoin(self.dir_master, 'metadata'))
        ensure_dirs(pjoin(self.dir_slave, 'metadata'))
        # silence missing masters warnings
        with open(pjoin(self.dir_master, 'metadata', 'layout.conf'), 'w') as f:
            f.write('masters =\n')
        with open(pjoin(self.dir_slave, 'metadata', 'layout.conf'), 'w') as f:
            f.write('masters = master\n')

        self.master_pdir = pjoin(self.dir_master, 'profiles')
        self.pdir = self.slave_pdir = pjoin(self.dir_slave, 'profiles')
        ensure_dirs(self.master_pdir)
        ensure_dirs(self.slave_pdir)
        # silence missing repo name warnings
        with open(pjoin(self.master_pdir, 'repo_name'), 'w') as f:
            f.write('master\n')
        with open(pjoin(self.slave_pdir, 'repo_name'), 'w') as f:
            f.write('slave\n')

        self.dir = self.dir_slave

    def tearDown(self):
        self.dir = self.dir_orig
        TempDirMixin.tearDown(self)

    def _create_categories(self, master, slave, write=True):
        with open(pjoin(self.master_pdir, 'categories'), 'w') as f:
            f.write('\n'.join(master))
        with open(pjoin(self.slave_pdir, 'categories'), 'w') as f:
            f.write('\n'.join(slave))
        for cat in master:
            os.mkdir(pjoin(self.dir_master, cat), 0o755)
        for cat in slave:
            os.mkdir(pjoin(self.dir_slave, cat), 0o755)
        return self.mk_tree(self.dir)

    def test_categories(self):
        # categories are inherited from masters
        for master, slave, expected in (
                (('cat',), (), ('cat',)),
                ((), ('cat',), ('cat',)),
                (('sys-apps', 'foo'), ('cat', 'foo'), ('cat', 'foo', 'sys-apps'))):
            # profiles/categories files exist along with category dirs
            self.setUp()
            repo = self._create_categories(master, slave)
            self.assertEqual(tuple(sorted(repo.categories)), expected)
            self.tearDown()

            # no profiles/categories files created, only category dirs
            self.setUp()
            repo = self._create_categories(master, slave, write=False)
            self.assertEqual(tuple(sorted(repo.categories)), expected)
            self.tearDown()
        self.setUp()

    def test_licenses(self):
        master_licenses = ('GPL-2', 'GPL-3+', 'BSD')
        slave_licenses = ('BSD-2', 'MIT')
        ensure_dirs(pjoin(self.dir_slave, 'licenses'))
        ensure_dirs(pjoin(self.dir_master, 'licenses'))
        for license in master_licenses:
            touch(pjoin(self.dir_master, 'licenses', license))
        for license in slave_licenses:
            touch(pjoin(self.dir_slave, 'licenses', license))
        repo = self.mk_tree(self.dir)
        self.assertEqual(sorted(repo.licenses), sorted(master_licenses + slave_licenses))

    def test_license_groups(self):
        master_licenses = ('GPL-2', 'BSD')
        slave_licenses = ('BSD-2', 'MIT')
        ensure_dirs(pjoin(self.dir_slave, 'licenses'))
        ensure_dirs(pjoin(self.dir_slave, 'profiles'))
        ensure_dirs(pjoin(self.dir_master, 'licenses'))
        ensure_dirs(pjoin(self.dir_master, 'profiles'))
        for license in master_licenses:
            touch(pjoin(self.dir_master, 'licenses', license))
        with open(pjoin(self.dir_master, 'profiles', 'license_groups'), 'w') as f:
            f.write(f'FREE {" ".join(master_licenses)}\n')
            f.write(f'OSI-APPROVED @FREE\n')
        for license in slave_licenses:
            touch(pjoin(self.dir_slave, 'licenses', license))
        with open(pjoin(self.dir_slave, 'profiles', 'license_groups'), 'w') as f:
            f.write(f'MISC-FREE @FREE {" ".join(slave_licenses)}\n')
            f.write('FSF-APPROVED MIT\n')
            f.write('OSI-APPROVED @FSF-APPROVED\n')
        repo = self.mk_tree(self.dir)
        self.assertEqual(sorted(repo.licenses), sorted(master_licenses + slave_licenses))
        self.assertEqual(sorted(repo.licenses.groups), ['FREE', 'FSF-APPROVED', 'MISC-FREE', 'OSI-APPROVED'])
        self.assertIn('BSD', repo.licenses.groups['MISC-FREE'])

    def test_package_deprecated(self):
        with open(pjoin(self.master_pdir, 'package.deprecated'), 'w') as f:
            f.write(textwrap.dedent('''\
                # lalala
                it-is/deprecated
                <just/newer-than-42
            '''))
        repo = self.mk_tree(self.dir_slave)
        self.assertEqual(sorted([atom('it-is/deprecated'),
            atom('<just/newer-than-42')]),
            sorted(list(repo.deprecated)))

    def test_use_expand_desc(self):
        use_expand_desc = {
            'example': (('example_foo', 'Build with foo'),
                        ('example_bar', 'Build with bar'))
        }
        ensure_dirs(pjoin(self.master_pdir, 'desc'))
        with open(pjoin(self.master_pdir, 'desc', 'example'), 'w') as f:
            f.write(textwrap.dedent('''\
                foo - Build with foo
                bar - Build with bar
            '''))
        repo = self.mk_tree(self.dir)
        assert use_expand_desc == dict(repo.use_expand_desc)

    def test_masters(self):
        repo = self.mk_tree(self.dir)
        self.assertEqual(repo.masters, (self.master_repo,))



# -*- coding: utf-8 -*-

__all__ = ("git_svn_syncer",)

import os

from . import base


class git_svn_syncer(base.VcsSyncer):

    binary = "git"

    supported_uris = (
        ('git+svn://', 10),
        ('git+svn+', 10),
    )

    @classmethod
    def is_usable_on_filepath(cls, path):
        git_svn_path = os.path.join(path, '.git', 'svn')
        if cls.disabled or not os.path.isdir(git_svn_path):
            return None
        return (cls._rewrite_uri_from_stat(git_svn_path, "git+svn://"),)

    @staticmethod
    def parse_uri(raw_uri):
        if not raw_uri.startswith("git+svn+") and not raw_uri.startswith("git+svn://"):
            raise base.UriError(
                raw_uri, "doesn't start with git+svn+ nor git+svn://")
        if raw_uri.startswith("git+svn+"):
            if raw_uri.startswith("git+svn+:"):
                raise base.UriError(
                    raw_uri, "need to specify the sub protocol if using git+svn+")
            return raw_uri[8:]
        return raw_uri[4:]

    def _initial_pull(self):
        return [self.binary_path, "svn", "clone", self.uri, self.basedir]

    def _update_existing(self):
        return [self.binary_path, "svn", "rebase"]



"""
errors fetch subsystem may throw
"""

from ..exceptions import PkgcoreUserException


class FetchError(PkgcoreUserException):
    """Generic fetch exception."""


class UnmodifiableFile(FetchError):

    def __init__(self, filename, extra=''):
        super().__init__(
            f'unable to update file {filename}, unmodifiable {extra}')
        self.filename = filename


class FetchFailed(FetchError):

    def __init__(self, filename, message, resumable=False):
        super().__init__(message)
        self.filename = filename
        self.message = message
        self.resumable = resumable

    def __str__(self):
        return f"failed fetching: {self.filename!r}: {self.message}"


class MissingDistfile(FetchFailed):

    def __init__(self, filename):
        super().__init__(filename, "doesn't exist", resumable=True)


class ChksumError(FetchError):
    """Generic checksum failure."""


class ChksumFailure(FetchFailed, ChksumError):
    """Checksum verification failed."""

    def __init__(self, filename, *, chksum, expected, value):
        self.filename = filename
        self.chksum = chksum
        self.expected = expected
        self.value = value
        super().__init__(filename, "checksum verification failed")


class RequiredChksumDataMissing(ChksumError):
    """A required checksum for the target is missing."""

    def __init__(self, fetchable, *chksum):
        super().__init__(
            f"chksum(s) {', '.join(chksum)} were configured as required, "
            f"but the data is missing: {fetchable.filename!r}"
        )
        self.fetchable, self.missing_chksum = fetchable, chksum


class MissingChksumHandler(ChksumError):
    """An unknown checksum type tried to be hashed."""
